# Copyright (C) 2023-2026 Jørgen Schartum Dokken
#
# This file is part of io4dolfinx
#
# SPDX-License-Identifier:    MIT


import inspect
import pathlib

from mpi4py import MPI

import dolfinx
import numpy as np
import pytest
import ufl

from io4dolfinx import (
    read_cell_data,
    read_function,
    read_function_from_legacy_h5,
    read_function_names,
    read_mesh,
    read_mesh_from_legacy_h5,
    read_meshtags,
    read_point_data,
)


def test_legacy_mesh(backend):
    comm = MPI.COMM_WORLD
    path = (pathlib.Path("legacy") / "mesh.h5").absolute()
    if not path.exists():
        pytest.skip(f"{path} does not exist")
    mesh = read_mesh_from_legacy_h5(filename=path, comm=comm, group="/mesh", backend=backend)
    assert mesh.topology.dim == 3
    volume = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.dx(domain=mesh))),
        op=MPI.SUM,
    )
    surface = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.ds(domain=mesh))),
        op=MPI.SUM,
    )
    assert np.isclose(volume, 1)
    assert np.isclose(surface, 6)

    mesh.topology.create_entities(mesh.topology.dim - 1)
    num_facets = mesh.topology.index_map(mesh.topology.dim - 1).size_global
    assert num_facets == 18


def test_read_legacy_mesh_from_checkpoint(backend):
    comm = MPI.COMM_WORLD
    filename = (pathlib.Path("legacy") / "mesh_checkpoint.h5").absolute()
    if not filename.exists():
        pytest.skip(f"{filename} does not exist")
    mesh = read_mesh_from_legacy_h5(
        filename=filename, comm=comm, group="/Mesh/mesh", backend=backend
    )
    assert mesh.topology.dim == 3
    volume = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.dx(domain=mesh))),
        op=MPI.SUM,
    )
    surface = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.ds(domain=mesh))),
        op=MPI.SUM,
    )
    assert np.isclose(volume, 1)
    assert np.isclose(surface, 6)

    mesh.topology.create_entities(mesh.topology.dim - 1)
    num_facets = mesh.topology.index_map(mesh.topology.dim - 1).size_global
    assert num_facets == 18


def test_legacy_function(backend):
    comm = MPI.COMM_WORLD
    path = (pathlib.Path("legacy") / "mesh.h5").absolute()
    if not path.exists():
        pytest.skip(f"{path} does not exist")
    mesh = read_mesh_from_legacy_h5(path, comm, "/mesh", backend=backend)
    ff = read_meshtags(
        path, mesh, "facet_tags", backend_args={"legacy": True, "engine": "HDF5"}, backend=backend
    )
    fdim = mesh.topology.dim - 1
    assert ff.dim == fdim
    boundary_facets = ff.find(2)
    interior_facets = ff.find(1)
    mesh.topology.create_connectivity(fdim, fdim + 1)
    num_facets_local = (
        mesh.topology.index_map(fdim).size_local + mesh.topology.index_map(fdim).num_ghosts
    )
    true_exterior_facets = dolfinx.mesh.exterior_facet_indices(mesh.topology)
    assert len(boundary_facets) == len(true_exterior_facets)
    assert np.isin(boundary_facets, true_exterior_facets).all()
    true_interior = np.ones(num_facets_local, dtype=np.int8)
    true_interior[true_exterior_facets] = 0
    interior_marker = np.flatnonzero(true_interior)
    assert len(interior_marker) == len(interior_facets)
    assert np.isin(interior_facets, interior_marker).all()

    V = dolfinx.fem.functionspace(mesh, ("DG", 2))
    u = ufl.TrialFunction(V)
    v = ufl.TestFunction(V)
    a = ufl.inner(u, v) * ufl.dx
    x = ufl.SpatialCoordinate(mesh)
    f = ufl.conditional(ufl.gt(x[0], 0.5), x[1], 2 * x[0])
    L = ufl.inner(f, v) * ufl.dx

    if not dolfinx.has_petsc4py:
        pytest.skip("dolfinx not configured with PETSc4py")

    from dolfinx.fem.petsc import LinearProblem

    uh = dolfinx.fem.Function(V)
    if "petsc_options_prefix" in inspect.signature(LinearProblem.__init__).parameters.keys():
        extra_options = {"petsc_options_prefix": "legacy_test"}
    else:
        extra_options = {}
    problem = LinearProblem(
        a, L, bcs=[], u=uh, petsc_options={"ksp_type": "preonly", "pc_type": "lu"}, **extra_options
    )
    problem.solve()

    u_in = dolfinx.fem.Function(V)
    read_function_from_legacy_h5(path, mesh.comm, u_in, group="v", backend=backend)
    np.testing.assert_allclose(uh.x.array, u_in.x.array, atol=1e-14)

    W = dolfinx.fem.functionspace(mesh, ("DG", 2, (mesh.geometry.dim,)))
    wh = dolfinx.fem.Function(W)
    wh.interpolate(lambda x: (x[0], 3 * x[2], 7 * x[1]))
    w_in = dolfinx.fem.Function(W)

    read_function_from_legacy_h5(path, mesh.comm, w_in, group="w", backend=backend)
    np.testing.assert_allclose(wh.x.array, w_in.x.array, atol=1e-14)


def test_read_legacy_function_from_checkpoint(backend):
    comm = MPI.COMM_WORLD
    path = (pathlib.Path("legacy") / "mesh_checkpoint.h5").absolute()
    if not path.exists():
        pytest.skip(f"{path} does not exist")

    from dolfinx.fem.petsc import LinearProblem

    mesh = read_mesh_from_legacy_h5(path, comm, "/Mesh/mesh", backend=backend)

    V = dolfinx.fem.functionspace(mesh, ("DG", 2))
    u = ufl.TrialFunction(V)
    v = ufl.TestFunction(V)
    a = ufl.inner(u, v) * ufl.dx
    x = ufl.SpatialCoordinate(mesh)
    f = ufl.conditional(ufl.gt(x[0], 0.5), x[1], 2 * x[0])
    L = ufl.inner(f, v) * ufl.dx

    if not dolfinx.has_petsc4py:
        pytest.skip("dolfinx not configured with PETSc4py")

    uh = dolfinx.fem.Function(V)
    if "petsc_options_prefix" in inspect.signature(LinearProblem.__init__).parameters.keys():
        extra_options = {"petsc_options_prefix": "legacy_checkpoint_test"}
    else:
        extra_options = {}
    problem = LinearProblem(
        a, L, bcs=[], u=uh, petsc_options={"ksp_type": "preonly", "pc_type": "lu"}, **extra_options
    )
    problem.solve()

    u_in = dolfinx.fem.Function(V)

    read_function_from_legacy_h5(path, mesh.comm, u_in, group="v", step=0, backend=backend)
    assert np.allclose(uh.x.array, u_in.x.array)

    # Check second step
    uh.interpolate(lambda x: x[0])
    read_function_from_legacy_h5(path, mesh.comm, u_in, group="v", step=1, backend=backend)
    assert np.allclose(uh.x.array, u_in.x.array)

    W = dolfinx.fem.functionspace(mesh, ("DG", 2, (mesh.geometry.dim,)))
    wh = dolfinx.fem.Function(W)
    wh.interpolate(lambda x: (x[0], 3 * x[2], 7 * x[1]))
    w_in = dolfinx.fem.Function(W)

    read_function_from_legacy_h5(path, mesh.comm, w_in, group="w", step=0, backend=backend)
    np.testing.assert_allclose(wh.x.array, w_in.x.array, atol=1e-14)

    wh.interpolate(lambda x: np.vstack((x[0], 0 * x[0], x[1])))
    read_function_from_legacy_h5(path, mesh.comm, w_in, group="w", step=1, backend=backend)
    np.testing.assert_allclose(wh.x.array, w_in.x.array, atol=1e-14)

    if backend == "h5py":
        qh = read_cell_data(
            path.with_suffix(".xdmf"), "vDG_checkpoint", mesh, 0.5, backend_args={}, backend="xdmf"
        )
        cell_map = mesh.topology.index_map(mesh.topology.dim)
        num_cells_local = cell_map.size_local + cell_map.num_ghosts
        midpoints = dolfinx.mesh.compute_midpoints(
            mesh, mesh.topology.dim, np.arange(num_cells_local, dtype=np.int32)
        )
        ref_val = midpoints[:, 0] + midpoints[:, 1]
        np.testing.assert_allclose(qh.x.array, ref_val)


def test_io4dolfinx_legacy():
    comm = MPI.COMM_WORLD
    path = (pathlib.Path("legacy_checkpoint") / "io4dolfinx_checkpoint.bp").absolute()
    if not path.exists():
        pytest.skip(f"{path} does not exist")

    el = ("N1curl", 3)
    backend_args = {"engine": "BP4", "legacy": True}
    mesh = read_mesh(
        path, comm, dolfinx.mesh.GhostMode.shared_facet, backend_args=backend_args, backend="adios2"
    )

    def f(x):
        values = np.zeros((2, x.shape[1]), dtype=np.float64)
        values[0] = x[0]
        values[1] = -x[1]
        return values

    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V)
    read_function(path, u, backend_args=backend_args, backend="adios2")

    u_ex = dolfinx.fem.Function(V)
    u_ex.interpolate(f)

    np.testing.assert_allclose(u.x.array, u_ex.x.array, atol=1e-14)


def test_legacy_vtu_mesh():
    comm = MPI.COMM_WORLD
    path = (pathlib.Path("legacy") / "mesh_P1000000.vtu").absolute()
    if not path.exists():
        pytest.skip(f"{path} does not exist")

    pytest.importorskip("pyvista")
    mesh = read_mesh(path, comm, backend="pyvista")

    num_cells_global = mesh.topology.index_map(mesh.topology.dim).size_global
    assert num_cells_global == 12 * 13 * 2

    area = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.dx(domain=mesh))), op=MPI.SUM
    )
    assert np.isclose(area, 1.9 * 2.8)


def test_legacy_pvd():
    comm = MPI.COMM_WORLD
    path = (pathlib.Path("legacy") / "mesh_P1_func.pvd").absolute()
    if not path.exists():
        pytest.skip(f"{path} does not exist")

    pytest.importorskip("pyvista")
    mesh = read_mesh(path, comm, backend="pyvista")

    num_cells_global = mesh.topology.index_map(mesh.topology.dim).size_global
    assert num_cells_global == 12 * 13 * 2

    area = mesh.comm.allreduce(
        dolfinx.fem.assemble_scalar(dolfinx.fem.form(1 * ufl.dx(domain=mesh))), op=MPI.SUM
    )
    assert np.isclose(area, 1.9 * 2.8)

    u = read_point_data(path, "w", mesh, backend="pyvista")
    names = read_function_names(path, mesh.comm, {}, backend="pyvista")
    assert len(names) == 1
    assert names[0] == "w"

    u_ref = dolfinx.fem.Function(u.function_space)
    u_ref.interpolate(lambda x: (x[0], x[1] - x[0], np.zeros_like(x[0])))

    np.testing.assert_allclose(u.x.array, u_ref.x.array)



# Copyright (C) 2023-2026 Jørgen Schartum Dokken
#
# This file is part of io4dolfinx
#
# SPDX-License-Identifier:    MIT

from __future__ import annotations

import typing
from pathlib import Path
from typing import Any

from mpi4py import MPI

import basix
import dolfinx
import numpy as np
import numpy.typing as npt
import ufl
from packaging.version import Version

from .backends import FileMode, ReadMode, get_backend
from .comm_helpers import (
    send_and_recv_cell_perm,
    send_dofmap_and_recv_values,
    send_dofs_and_recv_values,
)
from .readers import create_geometry_function_space
from .structures import ArrayData, FunctionData, MeshTagsData
from .utils import (
    check_file_exists,
    compute_dofmap_pos,
    compute_local_range,
    index_owner,
    unroll_dofmap,
    unroll_insert_position,
)
from .writers import prepare_meshdata_for_storage
from .writers import write_function as _internal_function_writer
from .writers import write_mesh as _internal_mesh_writer

__all__ = [
    "read_mesh",
    "write_function",
    "read_function",
    "write_mesh",
    "read_meshtags",
    "write_meshtags",
    "read_attributes",
    "write_attributes",
]


def write_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    attributes: dict[str, np.ndarray],
    backend_args: dict[str, typing.Any] | None = None,
    backend: str = "adios2",
):
    """Write attributes to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        name: Name of the attributes
        attributes: Dictionary of attributes to write to file
        backend_args: Arguments for backend, for instance file type.
        backend: What backend to use for writing.
    """
    backend_cls = get_backend(backend)
    backend_args = backend_cls.get_default_backend_args(backend_args)
    backend_cls.write_attributes(filename, comm, name, attributes, backend_args)


def read_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, typing.Any] | None = None,
    backend: str = "adios2",
) -> dict[str, typing.Any]:
    """Read attributes from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the attributes
        backend_args: Arguments for backend, for instance file type.
        backend: What backend to use for writing.
    Returns:
        The attributes
    """
    backend_cls = get_backend(backend)
    backend_args = backend_cls.get_default_backend_args(backend_args)
    return backend_cls.read_attributes(filename, comm, name, backend_args)


def read_timestamps(
    filename: Path | str,
    comm: MPI.Intracomm,
    function_name: str,
    backend_args: dict[str, typing.Any] | None = None,
    backend: str = "adios2",
) -> npt.NDArray[np.float64 | str]:  # type: ignore[type-var]
    """
    Read time-stamps from a checkpoint file.

    Args:
        comm: MPI communicator
        filename: Path to file
        function_name: Name of the function to read time-stamps for
        backend_args: Arguments for backend, for instance file type.
        backend: What backend to use for writing.
    Returns:
        The time-stamps
    """
    check_file_exists(filename)
    backend_cls = get_backend(backend)
    backend_args = backend_cls.get_default_backend_args(backend_args)
    return backend_cls.read_timestamps(filename, comm, function_name, backend_args)


def write_meshtags(
    filename: Path | str,
    mesh: dolfinx.mesh.Mesh,
    meshtags: dolfinx.mesh.MeshTags,
    meshtag_name: typing.Optional[str] = None,
    backend_args: dict[str, Any] | None = None,
    backend: str = "adios2",
):
    """
    Write meshtags associated with input mesh to file.

    .. note::
        For this checkpoint to work, the mesh must be written to file
        using :func:`write_mesh` before calling this function.

    Args:
        filename: Path to save meshtags (with file-extension)
        mesh: The mesh associated with the meshtags
        meshtags: The meshtags to write to file
        meshtag_name: Name of the meshtag. If None, the meshtag name is used.
        backend_args: Option to IO backend.
        backend: IO backend
    """

    # Extract data from meshtags (convert to global geometry node indices for each entity)
    tag_entities = meshtags.indices
    dim = meshtags.dim
    num_tag_entities_local = mesh.topology.index_map(dim).size_local
    local_tag_entities = tag_entities[tag_entities < num_tag_entities_local]
    local_values = meshtags.values[: len(local_tag_entities)]

    num_saved_tag_entities = len(local_tag_entities)
    local_start = mesh.comm.exscan(num_saved_tag_entities, op=MPI.SUM)
    local_start = local_start if mesh.comm.rank != 0 else 0
    global_num_tag_entities = mesh.comm.allreduce(num_saved_tag_entities, op=MPI.SUM)
    dof_layout = mesh.geometry.cmap.create_dof_layout()
    if hasattr(dof_layout, "num_entity_closure_dofs"):
        num_dofs_per_entity = dof_layout.num_entity_closure_dofs(dim)
    else:
        num_dofs_per_entity = len(dof_layout.entity_closure_dofs(dim, 0))

    entities_to_geometry = dolfinx.cpp.mesh.entities_to_geometry(
        mesh._cpp_object, dim, local_tag_entities, False
    )

    indices = (
        mesh.geometry.index_map()
        .local_to_global(entities_to_geometry.reshape(-1))
        .reshape(entities_to_geometry.shape)
    )
    name = meshtag_name or meshtags.name

    tag_ct = dolfinx.cpp.mesh.cell_entity_type(mesh.topology.cell_type, dim, 0).name
    tag_data = MeshTagsData(
        values=local_values,
        num_entities_global=global_num_tag_entities,
        num_dofs_per_entity=num_dofs_per_entity,
        indices=indices,
        name=name,
        local_start=local_start,
        dim=meshtags.dim,
        cell_type=tag_ct,
    )

    # Get backend and default arguments
    backend_cls = get_backend(backend)
    backend_args = backend_cls.get_default_backend_args(backend_args)
    return backend_cls.write_meshtags(filename, mesh.comm, tag_data, backend_args=backend_args)


def read_meshtags(
    filename: Path | str,
    mesh: dolfinx.mesh.Mesh,
    meshtag_name: str,
    backend_args: dict[str, Any] | None = None,
    backend: str = "adios2",
) -> dolfinx.mesh.MeshTags:
    """
    Read meshtags from file and return a :class:`dolfinx.mesh.MeshTags` object.

    Args:
        filename: Path to meshtags file (with file-extension)
        mesh: The mesh associated with the meshtags
        meshtag_name: The name of the meshtag to read
        engine: Adios2 Engine
    Returns:
        The meshtags
    """
    check_file_exists(filename)
    backend_cls = get_backend(backend)
    backend_args = backend_cls.get_default_backend_args(backend_args)
    data = backend_cls.read_meshtags_data(filename, mesh.comm, meshtag_name, backend_args)

    local_entities, local_values = dolfinx.io.distribute_entity_data(
        mesh, int(data.dim), data.indices, data.values
    )
    mesh.topology.create_connectivity(data.dim, 0)
    mesh.topology.create_connectivity(data.dim, mesh.topology.dim)

    adj = dolfinx.graph.adjacencylist(local_entities)

    local_values = np.array(local_values, dtype=np.int32)

    mt = dolfinx.mesh.meshtags_from_entities(mesh, int(data.dim), adj, local_values)
    mt.name = meshtag_name
    return mt


def read_function(
    filename: Path | str,
    u: dolfinx.fem.Function,
    time: float = 0.0,
    name: str | None = None,
    backend_args: dict[str, Any] | None = None,
    backend: str = "adios2",
):
    """
    Read checkpoint from file and fill it into `u`.

    Args:
        filename: Path to checkpoint
        u: Function to fill
        time: Time-stamp associated with checkpoint
        name: If not provided, `u.name` is used to search through the input file for the function
    """
    check_file_exists(filename)

    mesh = u.function_space.mesh
    comm = mesh.comm
    if name is None:
        name = u.name

    # ----------------------Step 1---------------------------------
    # Compute index of input cells and get cell permutation
    num_owned_cells = mesh.topology.index_map(mesh.topology.dim).size_local
    input_cells = mesh.topology.original_cell_index[:num_owned_cells]
    mesh.topology.create_entity_permutations()
    cell_perm = mesh.topology.get_cell_permutation_info()[:num_owned_cells]

    # Compute mesh->input communicator
    # 1.1 Compute mesh->input communicator
    backend_cls = get_backend(backend)
    owners: npt.NDArray[np.int32]
    if backend_cls.read_mode == ReadMode.serial:
        owners = np.zeros(input_cells, dtype=np.int32)
    elif backend_cls.read_mode == ReadMode.parallel:
        num_cells_global = mesh.topology.index_map(mesh.topology.dim).size_global
        owners = index_owner(mesh.comm, input_cells, num_cells_global)
    else:
        raise NotImplementedError(f"{backend_cls.read_mode} not implemented")
    # -------------------Step 2------------------------------------
    # Send and receive global cell index and cell perm
    inc_cells, inc_perms = send_and_recv_cell_perm(input_cells, cell_perm, owners, mesh.comm)

    # -------------------Step 3-----------------------------------
    # Read dofmap from file and compute dof owners
    check_file_exists(filename)
    backend_cls = get_backend(backend)
    backend_args = backend_cls.get_default_backend_args(backend_args)

    input_dofmap = backend_cls.read_dofmap(filename, comm, name, backend_args)

    # Compute owner of dofs in dofmap
    dof_owner: npt.NDArray[np.int32]
    if backend_cls.read_mode == ReadMode.serial:
        dof_owner = np.zeros(len(input_dofmap.array), dtype=np.int32)
    elif backend_cls.read_mode == ReadMode.parallel:
        num_dofs_global = (
            u.function_space.dofmap.index_map.size_global * u.function_space.dofmap.index_map_bs
        )
        dof_owner = index_owner(comm, input_dofmap.array.astype(np.int64), num_dofs_global)
    else:
        raise NotImplementedError(f"{backend_cls.read_mode} not implemented")

    # --------------------Step 4-----------------------------------
    # Read array from file and communicate them to input dofmap process
    input_array, starting_pos = backend_cls.read_dofs(filename, comm, name, time, backend_args)

    recv_array = send_dofs_and_recv_values(
        input_dofmap.array.astype(np.int64), dof_owner, comm, input_array, starting_pos
    )

    # -------------------Step 5--------------------------------------
    # Invert permutation of input data based on input perm
    # Then apply current permutation to the local data
    element = u.function_space.element
    if element.needs_dof_transformations:
        bs = u.function_space.dofmap.bs

        # Read input cell permutations on dofmap process
        local_input_range = compute_local_range(comm, num_cells_global)
        input_local_cell_index = inc_cells - local_input_range[0]
        input_perms = backend_cls.read_cell_perms(comm, filename, backend_args)

        # Start by sorting data array by cell permutation
        num_dofs_per_cell = input_dofmap.offsets[1:] - input_dofmap.offsets[:-1]
        assert np.allclose(num_dofs_per_cell, num_dofs_per_cell[0])

        # Sort dofmap by input local cell index
        input_perms_sorted = input_perms[input_local_cell_index]
        unrolled_dofmap_position = unroll_insert_position(
            input_local_cell_index, num_dofs_per_cell[0]
        )
        dofmap_sorted_by_input = recv_array[unrolled_dofmap_position]

        # First invert input data to reference element then transform to current mesh
        element.Tt_apply(dofmap_sorted_by_input, input_perms_sorted, bs)
        element.Tt_inv_apply(dofmap_sorted_by_input, inc_perms, bs)
        # Compute invert permutation
        inverted_perm = np.empty_like(unrolled_dofmap_position)
        inverted_perm[unrolled_dofmap_position] = np.arange(
            len(unrolled_dofmap_position), dtype=inverted_perm.dtype
        )
        recv_array = dofmap_sorted_by_input[inverted_perm]

    # ------------------Step 6----------------------------------------
    # For each dof owned by a process, find the local position in the dofmap.
    V = u.function_space
    local_cells, dof_pos = compute_dofmap_pos(V)
    input_cells = V.mesh.topology.original_cell_index[local_cells]
    num_cells_global = V.mesh.topology.index_map(V.mesh.topology.dim).size_global

    if backend_cls.read_mode == ReadMode.serial:
        owners = np.zeros(len(input_cells), dtype=np.int32)
    elif backend_cls.read_mode == ReadMode.parallel:
        owners = index_owner(V.mesh.comm, input_cells, num_cells_global)
    else:
        raise NotImplementedError(f"{backend_cls.read_mode} not implemented")

    unique_owners, owner_count = np.unique(owners, return_counts=True)
    # FIXME: In C++ use NBX to find neighbourhood
    sub_comm = V.mesh.comm.Create_dist_graph(
        [V.mesh.comm.rank], [len(unique_owners)], unique_owners, reorder=False
    )
    source, dest, _ = sub_comm.Get_dist_neighbors()
    sub_comm.Free()

    owned_values = send_dofmap_and_recv_values(
        comm,
        np.asarray(source, dtype=np.int32),
        np.asarray(dest, dtype=np.int32),
        owners,
        owner_count.astype(np.int32),
        input_cells,
        dof_pos,
        num_cells_global,
        recv_array,
        input_dofmap.offsets,
    )
    u.x.array[: len(owned_values)] = owned_values
    u.x.scatter_forward()


def read_mesh(
    filename: Path | str,
    comm: MPI.Intracomm,
    ghost_mode: dolfinx.mesh.GhostMode = dolfinx.mesh.GhostMode.shared_facet,
    time: float | str | None = 0.0,
    read_from_partition: bool = False,
    backend_args: dict[str, Any] | None = None,
    backend: str = "adios2",
    max_facet_to_cell_links: int = 2,
) -> dolfinx.mesh.Mesh:
    """
    Read an ADIOS2 mesh into DOLFINx.

    Args:
        filename: Path to input file
        comm: The MPI communciator to distribute the mesh over
        ghost_mode: Ghost mode to use for mesh. If `read_from_partition`
            is set to `True` this option is ignored.
        time: Time stamp associated with mesh
        read_from_partition: Read mesh with partition from file
        backend_args: List of arguments to reader backend
        max_facet_to_cell_links: Maximum number of cells a facet
            can be connected to.
    Returns:
        The distributed mesh
    """
    # Read in data in a distributed fashin
    check_file_exists(filename)
    backend_cls = get_backend(backend)
    backend_args = backend_cls.get_default_backend_args(backend_args)

    # Let each backend handle what should be default behavior when reading mesh
    # with or without time stamp.
    dist_in_data = backend_cls.read_mesh_data(
        filename,
        comm,
        time=time,
        read_from_partition=read_from_partition,
        backend_args=backend_args,
    )

    # Create DOLFINx mesh
    element = basix.ufl.element(
        basix.ElementFamily.P,
        dist_in_data.cell_type,
        dist_in_data.degree,
        basix.LagrangeVariant(int(dist_in_data.lvar)),
        shape=(dist_in_data.x.shape[1],),
        dtype=dist_in_data.x.dtype,
    )
    domain = ufl.Mesh(element)
    if (partition_graph := dist_in_data.partition_graph) is not None:

        def partitioner(comm: MPI.Intracomm, n, m, topo):
            assert len(topo[0]) % (len(partition_graph.offsets) - 1) == 0
            if Version(dolfinx.__version__) > Version("0.9.0"):
                return partition_graph._cpp_object
            else:
                return partition_graph
    else:
        try:
            partitioner = dolfinx.cpp.mesh.create_cell_partitioner(
                ghost_mode, max_facet_to_cell_links=max_facet_to_cell_links
            )
        except TypeError:
            partitioner = dolfinx.cpp.mesh.create_cell_partitioner(ghost_mode)

        # Should change to the commented code below when we require python
        # minimum version to be >=3.12 see https://github.com/python/cpython/pull/116198
        # import inspect
        # sig = inspect.signature(dolfinx.mesh.create_cell_partitioner)
        # part_kwargs = {}
        # if "max_facet_to_cell_links" in list(sig.parameters.keys()):
        #     part_kwargs["max_facet_to_cell_links"] = max_facet_to_cell_links
        # partitioner = dolfinx.cpp.mesh.create_cell_partitioner(ghost_mode, **part_kwargs)

    return dolfinx.mesh.create_mesh(
        comm,
        cells=dist_in_data.cells,
        x=dist_in_data.x,
        e=domain,
        partitioner=partitioner,
    )


def write_mesh(
    filename: Path,
    mesh: dolfinx.mesh.Mesh,
    mode: FileMode = FileMode.write,
    time: float = 0.0,
    store_partition_info: bool = False,
    backend_args: dict[str, Any] | None = None,
    backend: str = "adios2",
):
    """
    Write a mesh to file.

    Args:
        filename: Path to save mesh (without file-extension)
        mesh: The mesh to write to file

        store_partition_info: Store mesh partitioning (including ghosting) to file
    """
    mesh_data = prepare_meshdata_for_storage(mesh=mesh, store_partition_info=store_partition_info)

    _internal_mesh_writer(
        filename,
        mesh.comm,
        mesh_data=mesh_data,
        time=time,
        backend_args=backend_args,
        backend=backend,
        mode=mode,
    )


def write_function(
    filename: Path | str,
    u: dolfinx.fem.Function,
    time: float = 0.0,
    mode: FileMode = FileMode.append,
    name: str | None = None,
    backend_args: dict[str, Any] | None = None,
    backend: str = "adios2",
):
    """
    Write function checkpoint to file.

    Args:
        u: Function to write to file
        time: Time-stamp for simulation
        filename: Path to write to
        mode: Write or append.
        name: Name of function to write. If None, the name of the function is used.
        backend_args: Arguments to the IO backend.
        backend: The backend to use
    """
    dofmap = u.function_space.dofmap
    values = u.x.array
    mesh = u.function_space.mesh
    comm = mesh.comm
    mesh.topology.create_entity_permutations()
    cell_perm = mesh.topology.get_cell_permutation_info()
    num_cells_local = mesh.topology.index_map(mesh.topology.dim).size_local
    local_cell_range = mesh.topology.index_map(mesh.topology.dim).local_range
    num_cells_global = mesh.topology.index_map(mesh.topology.dim).size_global

    # Convert local dofmap into global_dofmap
    dmap = dofmap.list
    num_dofs_per_cell = dmap.shape[1]
    dofmap_bs = dofmap.bs
    num_dofs_local_dmap = num_cells_local * num_dofs_per_cell * dofmap_bs
    index_map_bs = dofmap.index_map_bs

    # Unroll dofmap for block size
    unrolled_dofmap = unroll_dofmap(dofmap.list[:num_cells_local, :], dofmap_bs)
    dmap_loc = (unrolled_dofmap // index_map_bs).reshape(-1)
    dmap_rem = (unrolled_dofmap % index_map_bs).reshape(-1)

    # Convert imap index to global index
    imap_global = dofmap.index_map.local_to_global(dmap_loc)
    dofmap_global = imap_global * index_map_bs + dmap_rem
    dofmap_imap = dolfinx.common.IndexMap(mesh.comm, num_dofs_local_dmap)

    # Compute dofmap offsets
    local_dofmap_offsets = np.arange(num_cells_local + 1, dtype=np.int64)
    local_dofmap_offsets[:] *= num_dofs_per_cell * dofmap_bs
    local_dofmap_offsets += dofmap_imap.local_range[0]

    num_dofs_global = dofmap.index_map.size_global * dofmap.index_map_bs
    local_dof_range = np.asarray(dofmap.index_map.local_range) * dofmap.index_map_bs
    num_dofs_local = local_dof_range[1] - local_dof_range[0]

    # Create internal data structure for function data to write to file
    function_data = FunctionData(
        cell_permutations=cell_perm[:num_cells_local].copy(),
        local_cell_range=local_cell_range,
        num_cells_global=num_cells_global,
        dofmap_array=dofmap_global,
        dofmap_offsets=local_dofmap_offsets,
        dofmap_range=dofmap_imap.local_range,
        global_dofs_in_dofmap=dofmap_imap.size_global,
        values=values[:num_dofs_local].copy(),
        dof_range=local_dof_range,
        num_dofs_global=num_dofs_global,
        name=name or u.name,
    )
    # Write to file
    fname = Path(filename)
    _internal_function_writer(
        fname, comm, function_data, time, backend_args=backend_args, backend=backend, mode=mode
    )


def read_function_names(
    filename: Path | str,
    comm: MPI.Intracomm,
    backend_args: dict[str, Any] | None = None,
    backend: str = "h5py",
) -> list[str]:
    """Read all function names from a file.

    Args:
        filename: Path to file
        comm: MPI communicator to launch IO on.
        backend_args: Arguments to backend

    Returns:
        A list of function names.
    """
    backend_cls = get_backend(backend)
    return backend_cls.read_function_names(filename, comm, backend_args=backend_args)


def write_point_data(
    filename: Path | str,
    u: dolfinx.fem.Function,
    time: str | float | None,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
    backend: str = "vtkhdf",
):
    """Write function to file by interpolating into geometry nodes.


    Args:
        filename: Path to file
        u: The function to store
        time: Time stamp
        mode: Append or write
        backend_args: The backend arguments
        backend: Which backend to use.
    """
    V = create_geometry_function_space(u.function_space.mesh, int(np.prod(u.ufl_shape)))
    v_out = dolfinx.fem.Function(V, name=u.name, dtype=u.x.array.dtype)
    v_out.interpolate(u)
    comm = v_out.function_space.mesh.comm
    data_shape = (V.dofmap.index_map.size_global, V.dofmap.index_map_bs)
    local_range = V.dofmap.index_map.local_range
    num_dofs_local = V.dofmap.index_map.size_local
    data = v_out.x.array.reshape(-1, V.dofmap.index_map_bs)[:num_dofs_local]
    ad = ArrayData(
        name=v_out.name, values=data, global_shape=data_shape, local_range=local_range, type="Point"
    )
    backend_cls = get_backend(backend)
    return backend_cls.write_data(
        filename, comm=comm, mode=mode, time=time, array_data=ad, backend_args=backend_args
    )


def write_cell_data(
    filename: Path | str,
    u: dolfinx.fem.Function,
    time: str | float | None,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
    backend: str = "vtkhdf",
):
    """Write function to file by interpolating into cell midpoints.


    Args:
        filename: Path to file
        point_data: Data to write to file
        time: Time stamp
        mode: Append or write
        backend_args: The backend arguments
    """
    V = dolfinx.fem.functionspace(u.function_space.mesh, ("DG", 0, u.ufl_shape))
    v_out = dolfinx.fem.Function(V, name=u.name, dtype=u.x.array.dtype)
    v_out.interpolate(u)
    comm = v_out.function_space.mesh.comm
    data_shape = (V.dofmap.index_map.size_global, V.dofmap.index_map_bs)
    local_range = V.dofmap.index_map.local_range
    num_dofs_local = V.dofmap.index_map.size_local
    data = v_out.x.array.reshape(-1, V.dofmap.index_map_bs)[:num_dofs_local]

    backend_cls = get_backend(backend)
    ad = ArrayData(
        name=v_out.name, values=data, global_shape=data_shape, local_range=local_range, type="Cell"
    )
    backend_cls = get_backend(backend)
    return backend_cls.write_data(
        filename, comm=comm, mode=mode, time=time, array_data=ad, backend_args=backend_args
    )



from mpi4py import MPI

import numpy as np
import pytest
import ufl
from dolfinx.fem import Function, assemble_scalar, form, functionspace
from dolfinx.io.vtkhdf import write_cell_data, write_mesh, write_point_data
from dolfinx.mesh import CellType, compute_midpoints, create_unit_cube, create_unit_square, meshtags

import io4dolfinx


def f(x, t):
    return x[0] - 2 * x[1] + x[2] * t


def g(x, t):
    return x[0], 2 * x[1], -x[2] * t


@pytest.mark.parametrize(
    "cell_type", [CellType.tetrahedron, CellType.hexahedron, CellType.tetrahedron]
)
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_read_write_timedep_mesh(dtype, tmp_path, cell_type):
    comm = MPI.COMM_WORLD
    tmp_path = comm.bcast(tmp_path, root=0)
    comm.barrier()

    # Write temporal data
    mesh = create_unit_cube(comm, 5, 5, 5, dtype=dtype, cell_type=cell_type)
    ref_vol = mesh.comm.allreduce(
        assemble_scalar(form(1 * ufl.dx(domain=mesh), dtype=dtype)), op=MPI.SUM
    )
    ref_surf = mesh.comm.allreduce(
        assemble_scalar(form(1 * ufl.ds(domain=mesh), dtype=dtype)), op=MPI.SUM
    )

    # Write temporal data
    filename = tmp_path / f"timedep_mesh_{cell_type.name}_{np.dtype(dtype).name}.vtkhdf"
    io4dolfinx.write_mesh(filename, mesh, time=0.3, backend="vtkhdf")
    mesh.geometry.x[:, 0] += 0.05 * mesh.geometry.x[:, 0]
    mesh.geometry.x[:, 1] *= 1.1 + np.sin(mesh.geometry.x[:, 0])

    ref_pert_vol = mesh.comm.allreduce(
        assemble_scalar(form(1 * ufl.dx(domain=mesh), dtype=dtype)), op=MPI.SUM
    )
    ref_pert_surf = mesh.comm.allreduce(
        assemble_scalar(form(1 * ufl.ds(domain=mesh), dtype=dtype)), op=MPI.SUM
    )

    io4dolfinx.write_mesh(
        filename,
        mesh,
        time=0.5,
        backend="vtkhdf",
        mode=io4dolfinx.FileMode.append,
    )

    in_mesh = io4dolfinx.read_mesh(filename, comm, time=0.5, backend="vtkhdf")
    pert_vol = mesh.comm.allreduce(
        assemble_scalar(form(1 * ufl.dx(domain=in_mesh), dtype=dtype)), op=MPI.SUM
    )
    pert_surf = mesh.comm.allreduce(
        assemble_scalar(form(1 * ufl.ds(domain=in_mesh), dtype=dtype)), op=MPI.SUM
    )
    assert np.isclose(pert_vol, ref_pert_vol)
    assert np.isclose(pert_surf, ref_pert_surf)

    in_mesh = io4dolfinx.read_mesh(filename, comm, time=0.3, backend="vtkhdf")
    vol = mesh.comm.allreduce(
        assemble_scalar(form(1 * ufl.dx(domain=in_mesh), dtype=dtype)), op=MPI.SUM
    )
    surf = mesh.comm.allreduce(
        assemble_scalar(form(1 * ufl.ds(domain=in_mesh), dtype=dtype)), op=MPI.SUM
    )
    assert np.isclose(vol, ref_vol)
    assert np.isclose(surf, ref_surf)


@pytest.mark.parametrize("cell_type", [CellType.hexahedron, CellType.tetrahedron])
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_write_point_data(dtype, tmp_path, cell_type):
    comm = MPI.COMM_WORLD
    tmp_path = comm.bcast(tmp_path, root=0)
    comm.barrier()

    # Write temporal data
    mesh = create_unit_cube(comm, 5, 5, 5, dtype=dtype, cell_type=cell_type)
    filename = tmp_path / f"point_data_{cell_type.name}_{np.dtype(dtype).name}.vtkhdf"
    write_mesh(str(filename), mesh)
    t = np.linspace(0.1, 1.2, 25)
    num_nodes_local = mesh.geometry.index_map().size_local
    for ti in t:
        point_data = f(mesh.geometry.x.T[:, :num_nodes_local], ti)
        write_point_data(str(filename), mesh, point_data, float(ti))
    comm.barrier()

    grid = io4dolfinx.read_mesh(filename=filename, comm=comm, time=None, backend="vtkhdf")
    # Since we shuffle time we need to shuffle in the same way on each process
    np.random.shuffle(t)
    t = comm.bcast(t, root=0)
    for tj in t:
        u = io4dolfinx.read_point_data(
            filename=filename, name="u", mesh=grid, time=tj, backend="vtkhdf"
        )
        v_ref = Function(u.function_space, dtype=u.x.array.dtype)
        atol = 15 * np.finfo(u.x.array.dtype).eps
        v_ref.interpolate(lambda x: f(x, tj))
        np.testing.assert_allclose(u.x.array, v_ref.x.array, atol=atol)

    # Test blocked data as well (with shuffled input timestep)
    blocked_file = filename.with_stem(filename.stem + "_blocked")
    write_mesh(str(blocked_file), mesh)
    for tj in t:
        point_data = np.asarray(g(mesh.geometry.x.T[:, :num_nodes_local], tj)).T.flatten()
        write_point_data(str(blocked_file), mesh, point_data, float(tj))
    comm.barrier()

    np.random.shuffle(t)
    t = comm.bcast(t, root=0)
    for tk in t:
        u = io4dolfinx.read_point_data(
            filename=blocked_file, name="u", mesh=grid, time=tk, backend="vtkhdf"
        )
        v_ref = Function(u.function_space, dtype=u.x.array.dtype)
        atol = 15 * np.finfo(u.x.array.dtype).eps
        v_ref.interpolate(lambda x: g(x, tk))
        np.testing.assert_allclose(u.x.array, v_ref.x.array, atol=atol)


@pytest.mark.parametrize("cell_type", [CellType.hexahedron, CellType.tetrahedron])
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_write_cell_data(dtype, tmp_path, cell_type):
    comm = MPI.COMM_WORLD
    tmp_path = comm.bcast(tmp_path, root=0)
    comm.barrier()

    # Write temporal data
    mesh = create_unit_cube(comm, 5, 5, 5, dtype=dtype, cell_type=cell_type)
    filename = tmp_path / f"cell_data_{cell_type.name}_{np.dtype(dtype).name}.vtkhdf"
    write_mesh(str(filename), mesh)

    t = np.linspace(0.1, 1.2, 25)
    num_cells_local = mesh.topology.index_map(mesh.topology.dim).size_local
    midpoints = compute_midpoints(
        mesh, mesh.topology.dim, np.arange(num_cells_local, dtype=np.int32)
    )
    for ti in t:
        cell_data = f(midpoints.T, ti)
        write_cell_data(str(filename), mesh, cell_data, float(ti))
    comm.barrier()

    grid = io4dolfinx.read_mesh(filename=filename, comm=comm, time=None, backend="vtkhdf")
    # Since we shuffle time we need to shuffle in the same way on each process
    np.random.shuffle(t)
    t = comm.bcast(t, root=0)
    for tj in t:
        u = io4dolfinx.read_cell_data(
            filename=filename, name="u", mesh=grid, time=tj, backend="vtkhdf"
        )
        v_ref = Function(u.function_space, dtype=u.x.array.dtype)
        atol = 15 * np.finfo(u.x.array.dtype).eps
        v_ref.interpolate(lambda x: f(x, tj))
        np.testing.assert_allclose(u.x.array, v_ref.x.array, atol=atol)

    # Test blocked data as well (with shuffled input timestep)
    blocked_file = filename.with_stem(filename.stem + "_blocked")
    write_mesh(str(blocked_file), mesh)
    for tj in t:
        cell_data = np.asarray(g(midpoints.T, tj)).T.flatten()
        write_cell_data(str(blocked_file), mesh, cell_data, float(tj))
    comm.barrier()

    np.random.shuffle(t)
    t = comm.bcast(t, root=0)
    for tk in t:
        u = io4dolfinx.read_cell_data(
            filename=blocked_file, name="u", mesh=grid, time=tk, backend="vtkhdf"
        )
        v_ref = Function(u.function_space, dtype=u.x.array.dtype)
        atol = 15 * np.finfo(u.x.array.dtype).eps
        v_ref.interpolate(lambda x: g(x, tk))
        np.testing.assert_allclose(u.x.array, v_ref.x.array, atol=atol)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_write_meshtags(dtype, tmp_path, generate_reference_map):
    comm = MPI.COMM_WORLD
    tmp_path = comm.bcast(tmp_path, root=0)
    comm.barrier()

    filename = tmp_path / f"meshtags_{np.dtype(dtype).name}.vtkhdf"

    mesh = create_unit_cube(comm, 3, 3, 3, dtype=dtype, cell_type=CellType.hexahedron)
    io4dolfinx.write_mesh(
        filename,
        mesh,
        mode=io4dolfinx.FileMode.write,
        time=1.0,
        backend_args={"name": "hex"},
        backend="vtkhdf",
    )
    dim = mesh.topology.dim
    mesh.topology.create_connectivity(dim, mesh.topology.dim)
    cmap = mesh.topology.index_map(dim)
    cells = np.arange(cmap.size_local, dtype=np.int32)
    ct = meshtags(mesh, dim, cells, cells + cmap.local_range[0])
    io4dolfinx.write_meshtags(
        filename, mesh, ct, "CellTags", backend_args={"name": "hex"}, backend="vtkhdf"
    )
    root = 0

    org_map = generate_reference_map(mesh, ct, comm, root)
    # Move mesh
    mesh.geometry.x[:, 0] *= 1 + 0.2 * np.sin(mesh.geometry.x[:, 1])

    io4dolfinx.write_mesh(
        filename,
        mesh,
        mode=io4dolfinx.FileMode.append,
        time=2.5,
        backend_args={"name": "hex"},
        backend="vtkhdf",
    )

    # Add stationary meshtags (after time loop)
    mesh = create_unit_cube(comm, 7, 3, 5, dtype=dtype, cell_type=CellType.tetrahedron)
    io4dolfinx.write_mesh(
        filename,
        mesh,
        mode=io4dolfinx.FileMode.append,
        time=1.0,
        backend_args={"name": "tet"},
        backend="vtkhdf",
    )
    dim = mesh.topology.dim
    org_maps = {}
    mesh.geometry.x[:, 0] *= 2.0 + mesh.geometry.x[:, 1]
    io4dolfinx.write_mesh(
        filename,
        mesh,
        mode=io4dolfinx.FileMode.append,
        time=2.5,
        backend_args={"name": "tet"},
        backend="vtkhdf",
    )
    for dim in range(mesh.topology.dim + 1):
        mesh.topology.create_connectivity(dim, mesh.topology.dim)
        entities = np.arange(mesh.topology.index_map(dim).size_local, dtype=np.int32)
        et = meshtags(mesh, dim, entities, entities + mesh.topology.index_map(dim).local_range[0])
        io4dolfinx.write_meshtags(
            filename, mesh, et, f"{dim}tags", backend_args={"name": "tet"}, backend="vtkhdf"
        )
        org_maps[dim] = generate_reference_map(mesh, et, comm, root)

    tol = 15 * np.finfo(dtype).eps
    # Read in hex grid from second time step
    hex_mesh = io4dolfinx.read_mesh(
        filename, comm, time=1.0, backend_args={"name": "hex"}, backend="vtkhdf"
    )
    hex_tag = io4dolfinx.read_meshtags(
        filename, hex_mesh, "CellTags", backend_args={"name": "hex"}, backend="vtkhdf"
    )
    read_map = generate_reference_map(hex_mesh, hex_tag, comm, root)
    # On root process, check that midpoints are the same for each value in the meshtag
    if MPI.COMM_WORLD.rank == root:
        assert len(org_map) == len(read_map)
        for value, (_, midpoint) in org_map.items():
            _, read_midpoint = read_map[value]
            np.testing.assert_allclose(read_midpoint, midpoint, atol=tol)

    # Read tet grid from second time step
    tet_mesh = io4dolfinx.read_mesh(
        filename, comm, time=2.5, backend_args={"name": "tet"}, backend="vtkhdf"
    )
    for dim in range(mesh.topology.dim + 1):
        tet_tag = io4dolfinx.read_meshtags(
            filename, tet_mesh, f"{dim}tags", backend_args={"name": "tet"}, backend="vtkhdf"
        )
        read_map = generate_reference_map(tet_mesh, tet_tag, comm, root)
        if MPI.COMM_WORLD.rank == root:
            assert len(org_maps[dim]) == len(read_map)
            for value, (_, midpoint) in org_maps[dim].items():
                _, read_midpoint = read_map[value]
                np.testing.assert_allclose(read_midpoint, midpoint, atol=tol)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_read_write_pointdata(dtype, tmp_path):
    tol = 15 * np.finfo(dtype).eps

    comm = MPI.COMM_WORLD
    tmp_path = comm.bcast(tmp_path, root=0)
    comm.barrier()

    filename = tmp_path / "point_data.vtkhdf"

    mesh = create_unit_cube(comm, 3, 3, 3, dtype=dtype, cell_type=CellType.hexahedron)

    def f(x, t):
        return (x[0] + np.sin(x[1]) + np.cos(x[0] * t), x[2] + x[1] - t)

    io4dolfinx.write_mesh(
        filename,
        mesh,
        mode=io4dolfinx.FileMode.write,
        time=1.0,
        backend_args={"name": "hex"},
        backend="vtkhdf",
    )

    f_name = "point_data"
    V = functionspace(mesh, ("Lagrange", 2, (2,)))
    u = Function(V, dtype=dtype, name=f_name)
    u.interpolate(lambda x: f(x, 1.0))
    io4dolfinx.write_point_data(
        filename,
        u,
        mode=io4dolfinx.FileMode.append,
        time=1.0,
        backend_args={"name": "hex"},
        backend="vtkhdf",
    )

    io4dolfinx.write_mesh(
        filename,
        mesh,
        mode=io4dolfinx.FileMode.append,
        time=2.0,
        backend_args={"name": "hex"},
        backend="vtkhdf",
    )
    u.interpolate(lambda x: f(x, 2.0))
    io4dolfinx.write_point_data(
        filename,
        u,
        mode=io4dolfinx.FileMode.append,
        time=2.0,
        backend_args={"name": "hex"},
        backend="vtkhdf",
    )

    # Read in hex grid from second time step
    hex_mesh = io4dolfinx.read_mesh(
        filename, comm, time=2.0, backend_args={"name": "hex"}, backend="vtkhdf"
    )
    u_end = io4dolfinx.read_point_data(
        filename,
        mesh=hex_mesh,
        name=f_name,
        time=2.0,
        backend_args={"name": "hex"},
        backend="vtkhdf",
    )
    u_ref = Function(u_end.function_space, dtype=dtype)
    u_ref.interpolate(lambda x: f(x, 2.0))
    np.testing.assert_allclose(u_end.x.array, u_ref.x.array, atol=tol)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_read_write_celldata(dtype, tmp_path):
    tol = 15 * np.finfo(dtype).eps

    comm = MPI.COMM_WORLD
    tmp_path = comm.bcast(tmp_path, root=0)
    comm.barrier()

    filename = tmp_path / "cell_data.vtkhdf"

    mesh = create_unit_cube(comm, 3, 3, 3, dtype=dtype, cell_type=CellType.tetrahedron)

    def f(x, t):
        return (x[0] + np.sin(x[1]) + np.cos(x[0] * t), x[1] + t, x[0] - t)

    t_0 = 2.2
    t_1 = 3.0

    backend_args = {"name": "Grid"}
    io4dolfinx.write_mesh(
        filename,
        mesh,
        mode=io4dolfinx.FileMode.write,
        time=t_0,
        backend_args=backend_args,
        backend="vtkhdf",
    )

    f_name = "Data"
    V = functionspace(mesh, ("DG", 0, (3,)))
    u = Function(V, dtype=dtype, name=f_name)
    u.interpolate(lambda x: f(x, t_0))
    io4dolfinx.write_cell_data(
        filename,
        u,
        mode=io4dolfinx.FileMode.append,
        time=t_0,
        backend_args=backend_args,
        backend="vtkhdf",
    )

    io4dolfinx.write_mesh(
        filename,
        mesh,
        mode=io4dolfinx.FileMode.append,
        time=t_1,
        backend_args=backend_args,
        backend="vtkhdf",
    )
    u.interpolate(lambda x: f(x, t_1))
    io4dolfinx.write_cell_data(
        filename,
        u,
        mode=io4dolfinx.FileMode.append,
        time=t_1,
        backend_args=backend_args,
        backend="vtkhdf",
    )

    for t in [t_1, t_0]:
        grid = io4dolfinx.read_mesh(
            filename, comm, time=t, backend_args=backend_args, backend="vtkhdf"
        )
        u_end = io4dolfinx.read_cell_data(
            filename,
            mesh=grid,
            name=f_name,
            time=t,
            backend_args=backend_args,
            backend="vtkhdf",
        )
        u_ref = Function(u_end.function_space, dtype=dtype)
        u_ref.interpolate(lambda x: f(x, t))
        np.testing.assert_allclose(u_end.x.array, u_ref.x.array, atol=tol)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_read_write_mix_data(dtype, tmp_path):
    tol = 15 * np.finfo(dtype).eps

    mesh = create_unit_square(MPI.COMM_WORLD, 5, 7, dtype=dtype)

    def f(x, t):
        return x[0] + t * x[1]

    def g(x, t):
        return x[1] - x[0] * t**2

    ts = [0.1, 0.3, 0.4, 0.5]

    V = functionspace(mesh, ("Lagrange", 1))
    u = Function(V, name="points", dtype=dtype)
    z = Function(V, name="some_other_array", dtype=dtype)
    Q = functionspace(mesh, ("DG", 0))
    q = Function(Q, name="cells", dtype=dtype)
    tmp_path = mesh.comm.bcast(tmp_path, root=0)
    filename = tmp_path / "mixed_data.vtkhdf"

    backend_args = {"name": "MyGrid"}
    for i, t in enumerate(ts):
        if np.isclose(t, 0.1):
            mode = io4dolfinx.FileMode.write
        else:
            mode = io4dolfinx.FileMode.append
        mesh.geometry.x[:] *= 1 + 0.1 * t
        io4dolfinx.write_mesh(
            filename, mesh, mode=mode, time=t, backend="vtkhdf", backend_args=backend_args
        )
        u.interpolate(lambda x: f(x, t))
        q.interpolate(lambda x: g(x, t))
        io4dolfinx.write_point_data(
            filename,
            u,
            time=t,
            mode=io4dolfinx.FileMode.append,
            backend_args=backend_args,
            backend="vtkhdf",
        )
        io4dolfinx.write_point_data(
            filename,
            z,
            time=t,
            mode=io4dolfinx.FileMode.append,
            backend_args=backend_args,
            backend="vtkhdf",
        )

        mesh_in = io4dolfinx.read_mesh(
            filename, MPI.COMM_WORLD, time=t, backend_args=backend_args, backend="vtkhdf"
        )
        u_in = io4dolfinx.read_point_data(
            filename, name=u.name, mesh=mesh_in, time=t, backend_args=backend_args, backend="vtkhdf"
        )
        u_ref = Function(u_in.function_space, dtype=dtype)
        u_ref.interpolate(lambda x: f(x, t))
        np.testing.assert_allclose(u_ref.x.array, u_in.x.array, atol=tol)
        c_step = i
        if not np.isclose(t, 0.3):
            io4dolfinx.write_cell_data(
                filename,
                q,
                time=t,
                backend_args=backend_args,
                mode=io4dolfinx.FileMode.append,
                backend="vtkhdf",
            )
        else:
            # Read in mesh from previous step as geometry adapts,
            # while reading data from current step.
            c_step = i - 1
            mesh_in = io4dolfinx.read_mesh(
                filename,
                MPI.COMM_WORLD,
                time=ts[c_step],
                backend_args=backend_args,
                backend="vtkhdf",
            )
        q_in = io4dolfinx.read_cell_data(
            filename, name=q.name, mesh=mesh_in, time=t, backend_args=backend_args, backend="vtkhdf"
        )
        q_ref = Function(q_in.function_space, dtype=dtype)
        q_ref.interpolate(lambda x: g(x, ts[c_step]))
        np.testing.assert_allclose(q_ref.x.array, q_in.x.array, atol=tol)



from pathlib import Path

from mpi4py import MPI

import basix.ufl
import dolfinx
import numpy as np
import pytest

from io4dolfinx import FileMode, snapshot_checkpoint


def suffix(backend: str) -> str:
    if backend == "adios2":
        return ".bp"
    elif backend == "h5py":
        return ".h5"
    else:
        raise NotImplementedError(f"Unsupported backend {backend}")


triangle = dolfinx.mesh.CellType.triangle
quad = dolfinx.mesh.CellType.quadrilateral
tetra = dolfinx.mesh.CellType.tetrahedron
hex = dolfinx.mesh.CellType.hexahedron


@pytest.mark.parametrize(
    "cell_type, family", [(triangle, "N1curl"), (triangle, "RT"), (quad, "RTCF")]
)
@pytest.mark.parametrize("degree", [1, 4])
def test_read_write_2D(family, degree, cell_type, tmp_path, backend):
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10, cell_type=cell_type)
    el = basix.ufl.element(family, mesh.basix_cell(), degree)

    def f(x):
        return (np.full(x.shape[1], np.pi) + x[0], x[1])

    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V)
    u.interpolate(f)

    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    file = fname / Path("snapshot_2D_vs").with_suffix(suffix(backend))
    snapshot_checkpoint(u, file, FileMode.write, backend=backend)

    v = dolfinx.fem.Function(V)
    snapshot_checkpoint(v, file, FileMode.read, backend=backend)
    assert np.allclose(u.x.array, v.x.array)


@pytest.mark.parametrize("cell_type, family", [(tetra, "N1curl"), (tetra, "RT"), (hex, "NCF")])
@pytest.mark.parametrize("degree", [1, 4])
def test_read_write_3D(family, degree, cell_type, tmp_path, backend):
    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 3, 3, 3, cell_type=cell_type)
    el = basix.ufl.element(family, mesh.basix_cell(), degree)

    def f(x):
        return (np.full(x.shape[1], np.pi) + x[0], x[1], x[1] * x[2])

    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V)
    u.interpolate(f)

    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    file = fname / Path("snapshot_3D_vs").with_suffix(suffix(backend))
    snapshot_checkpoint(u, file, FileMode.write, backend=backend)

    v = dolfinx.fem.Function(V)
    snapshot_checkpoint(v, file, FileMode.read, backend=backend)
    assert np.allclose(u.x.array, v.x.array)


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
@pytest.mark.parametrize("family", ["Lagrange", "DG"])
@pytest.mark.parametrize("degree", [1, 4])
def test_read_write_P_2D(family, degree, cell_type, tmp_path, backend):
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 5, 5, cell_type=cell_type)
    el = basix.ufl.element(family, mesh.basix_cell(), degree, shape=(mesh.geometry.dim,))

    def f(x):
        return (np.full(x.shape[1], np.pi) + x[0], x[1])

    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V)
    u.interpolate(f)

    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    file = fname / Path("snapshot_2D_p").with_suffix(suffix(backend))
    snapshot_checkpoint(u, file, FileMode.write, backend=backend)

    v = dolfinx.fem.Function(V)
    snapshot_checkpoint(v, file, FileMode.read, backend=backend)
    assert np.allclose(u.x.array, v.x.array)


@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
@pytest.mark.parametrize("family", ["Lagrange", "DG"])
@pytest.mark.parametrize("degree", [1, 4])
def test_read_write_P_3D(family, degree, cell_type, tmp_path, backend):
    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 5, 5, 5, cell_type=cell_type)
    el = basix.ufl.element(family, mesh.basix_cell(), degree, shape=(mesh.geometry.dim,))

    def f(x):
        return (np.full(x.shape[1], np.pi) + x[0], x[1] + 2 * x[0], np.cos(x[2]))

    V = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(V)
    u.interpolate(f)

    fname = MPI.COMM_WORLD.bcast(tmp_path, root=0)
    file = fname / Path("snapshot_3D_p").with_suffix(suffix(backend))

    snapshot_checkpoint(u, file, FileMode.write, backend=backend)

    v = dolfinx.fem.Function(V)
    snapshot_checkpoint(v, file, FileMode.read, backend=backend)
    assert np.allclose(u.x.array, v.x.array)



import itertools

from mpi4py import MPI

import basix
import basix.ufl
import dolfinx
import numpy as np
import pytest

dtypes = [np.float64, np.float32]  # Mesh geometry dtypes
write_comm = [MPI.COMM_SELF, MPI.COMM_WORLD]  # Communicators for creating mesh


simplex_two_dim = itertools.product(dtypes, [dolfinx.mesh.CellType.triangle], write_comm)
simplex_three_dim = itertools.product(dtypes, [dolfinx.mesh.CellType.tetrahedron], write_comm)

non_simplex_two_dim = itertools.product(dtypes, [dolfinx.mesh.CellType.quadrilateral], write_comm)
non_simplex_three_dim = itertools.product(dtypes, [dolfinx.mesh.CellType.hexahedron], write_comm)


@pytest.fixture(params=simplex_two_dim, scope="module")
def simplex_mesh_2D(request):
    dtype, cell_type, write_comm = request.param
    mesh = dolfinx.mesh.create_unit_square(write_comm, 10, 10, cell_type=cell_type, dtype=dtype)
    return mesh


@pytest.fixture(params=simplex_three_dim, scope="module")
def simplex_mesh_3D(request):
    dtype, cell_type, write_comm = request.param
    mesh = dolfinx.mesh.create_unit_cube(write_comm, 5, 5, 5, cell_type=cell_type, dtype=dtype)
    return mesh


@pytest.fixture(params=non_simplex_two_dim, scope="module")
def non_simplex_mesh_2D(request):
    dtype, cell_type, write_comm = request.param
    mesh = dolfinx.mesh.create_unit_square(write_comm, 10, 10, cell_type=cell_type, dtype=dtype)
    return mesh


@pytest.fixture(params=non_simplex_three_dim, scope="module")
def non_simplex_mesh_3D(request):
    dtype, cell_type, write_comm = request.param
    mesh = dolfinx.mesh.create_unit_cube(write_comm, 5, 5, 5, cell_type=cell_type, dtype=dtype)
    return mesh


@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["N1curl", "RT"])
@pytest.mark.parametrize("degree", [1, 4])
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_read_write_2D(
    read_comm,
    family,
    degree,
    is_complex,
    simplex_mesh_2D,
    get_dtype,
    write_function,
    read_function,
    backend,
):
    mesh = simplex_mesh_2D
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)
    el = basix.ufl.element(family, mesh.basix_cell(), degree, dtype=mesh.geometry.x.dtype)

    def f(x):
        values = np.empty((2, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi) + x[0]
        values[1] = x[1]
        if is_complex:
            values[0] += 2j * x[1]
            values[1] += 2j * x[0]
        return values

    fname = write_function(mesh, el, f, f_dtype, backend=backend)
    MPI.COMM_WORLD.Barrier()
    read_function(read_comm, el, f, fname, f_dtype, backend=backend)


@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["N1curl", "RT"])
@pytest.mark.parametrize("degree", [1, 4])
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_read_write_3D(
    read_comm,
    family,
    degree,
    is_complex,
    simplex_mesh_3D,
    get_dtype,
    write_function,
    read_function,
    backend,
):
    mesh = simplex_mesh_3D
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)
    el = basix.ufl.element(family, mesh.basix_cell(), degree, dtype=mesh.geometry.x.dtype)

    def f(x):
        values = np.empty((3, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi)
        values[1] = x[1] + 2 * x[0]
        values[2] = np.cos(x[2])
        if is_complex:
            values[0] += 2j * x[2]
            values[1] += 2j * np.cos(x[2])
        return values

    fname = write_function(mesh, el, f, dtype=f_dtype, backend=backend)
    MPI.COMM_WORLD.Barrier()
    read_function(read_comm, el, f, fname, dtype=f_dtype, backend=backend)


@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["RTCF"])
@pytest.mark.parametrize("degree", [1, 2, 3])
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_read_write_2D_quad(
    read_comm,
    family,
    degree,
    is_complex,
    non_simplex_mesh_2D,
    get_dtype,
    write_function,
    read_function,
    backend,
):
    mesh = non_simplex_mesh_2D
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)
    el = basix.ufl.element(family, mesh.basix_cell(), degree, dtype=mesh.geometry.x.dtype)

    def f(x):
        values = np.empty((2, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi)
        values[1] = x[1] + 2 * x[0]
        if is_complex:
            values[0] += 2j * x[2]
            values[1] += 2j * np.cos(x[2])
        return values

    hash = write_function(mesh, el, f, f_dtype, backend=backend)
    MPI.COMM_WORLD.Barrier()
    read_function(read_comm, el, f, hash, f_dtype, backend=backend)


@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["NCF"])
@pytest.mark.parametrize("degree", [1, 4])
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_read_write_hex(
    read_comm,
    family,
    degree,
    is_complex,
    non_simplex_mesh_3D,
    get_dtype,
    write_function,
    read_function,
    backend,
):
    mesh = non_simplex_mesh_3D
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)
    el = basix.ufl.element(family, mesh.basix_cell(), degree, dtype=mesh.geometry.x.dtype)

    def f(x):
        values = np.empty((3, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi) + x[0]
        values[1] = np.cos(x[2])
        values[2] = x[0]
        if is_complex:
            values[0] += 2j * x[2]
            values[2] -= 1j * x[1]
        return values

    hash = write_function(mesh, el, f, dtype=f_dtype, backend=backend)
    MPI.COMM_WORLD.Barrier()
    read_function(read_comm, el, f, hash, dtype=f_dtype, backend=backend)


@pytest.mark.parametrize("is_complex", [True, False])
@pytest.mark.parametrize("family", ["RTCF"])
@pytest.mark.parametrize("degree", [1, 2, 3])
@pytest.mark.parametrize("read_comm", [MPI.COMM_SELF, MPI.COMM_WORLD])
def test_read_write_multiple(
    read_comm,
    family,
    degree,
    is_complex,
    non_simplex_mesh_2D,
    get_dtype,
    write_function,
    read_function,
    backend,
):
    mesh = non_simplex_mesh_2D
    f_dtype = get_dtype(mesh.geometry.x.dtype, is_complex)
    el = basix.ufl.element(family, mesh.basix_cell(), degree, dtype=mesh.geometry.x.dtype)

    def f(x):
        values = np.empty((2, x.shape[1]), dtype=f_dtype)
        values[0] = np.full(x.shape[1], np.pi)
        values[1] = x[1] + 2 * x[0]
        if is_complex:
            values[0] -= 2j * x[2]
            values[1] += 2j * np.cos(x[2])
        return values

    def g(x):
        values = np.empty((2, x.shape[1]), dtype=f_dtype)
        values[0] = 2 * x[1]
        values[1] = 3 * x[0]
        if is_complex:
            values[0] += 3j * x[0]
            values[1] += 2j * x[0] * x[1]
        return values

    hash_f = write_function(mesh, el, f, dtype=f_dtype, name="f", append=False, backend=backend)
    hash_g = write_function(mesh, el, g, dtype=f_dtype, name="g", append=True, backend=backend)
    assert hash_f == hash_g

    MPI.COMM_WORLD.Barrier()
    read_function(read_comm, el, f, hash_f, dtype=f_dtype, name="f", backend=backend)
    read_function(read_comm, el, g, hash_g, dtype=f_dtype, name="g", backend=backend)



from __future__ import annotations

from mpi4py import MPI

import numpy as np
import numpy.typing as npt

from .utils import compute_insert_position, compute_local_range, valid_function_types

__all__ = [
    "send_dofmap_and_recv_values",
    "send_and_recv_cell_perm",
    "send_dofs_and_recv_values",
    "numpy_to_mpi",
]
"""
Helpers for sending and receiving values for checkpointing
"""

numpy_to_mpi = {
    np.float64: MPI.DOUBLE,
    np.float32: MPI.FLOAT,
    np.complex64: MPI.COMPLEX,
    np.complex128: MPI.DOUBLE_COMPLEX,
    np.int64: MPI.INT64_T,
    np.int32: MPI.INT32_T,
}


def send_dofmap_and_recv_values(
    comm: MPI.Intracomm,
    source_ranks: npt.NDArray[np.int32],
    dest_ranks: npt.NDArray[np.int32],
    output_owners: npt.NDArray[np.int32],
    dest_size: npt.NDArray[np.int32],
    input_cells: npt.NDArray[np.int64],
    dofmap_pos: npt.NDArray[np.int32],
    num_cells_global: np.int64,
    values: npt.NDArray[valid_function_types],
    dofmap_offsets: npt.NDArray[np.int32],
) -> npt.NDArray[valid_function_types]:
    """
    Given a set of positions in input dofmap, give the global input index of this dofmap entry
    in input file.

    Args:
        comm: The MPI communicator to create the Neighbourhood-communicator from
        source_ranks: Ranks that will send dofmap indices to current process
        dest_ranks: Ranks that will receive dofmap indices from current process
        output_owners: The owners of each dofmap entry on this process. The unique set of
            these entries should be the same as the dest_ranks.
        dest_size: The number of entries sent to each owner
        input_cells: A cell associated with the degree of freedom sent (global index).
        dofmap_pos: The local position in the dofmap. I.e.
            `dof = dofmap.links(input_cells)[dofmap_pos]`
        num_cells_global: Number of global cells
        values: Values currently held by this process. These are
            ordered (num_cells_local, num_dofs_per_cell), flattened row-major.
        dofmap_offsets: Local dofmap offsets to access the correct `values`.

    Returns:
        Values corresponding to the dofs owned by this process.
    """
    insert_position = compute_insert_position(output_owners, dest_ranks, dest_size)

    # Pack the cells and dofmap position for all dofs this process is distributing
    out_cells = np.zeros(len(output_owners), dtype=np.int64)
    out_cells[insert_position] = input_cells
    out_pos = np.zeros(len(output_owners), dtype=np.int32)
    out_pos[insert_position] = dofmap_pos

    # Compute map from the data index sent to each process and the local
    # number on the current process
    proc_to_dof = np.zeros_like(input_cells, dtype=np.int32)
    proc_to_dof[insert_position] = np.arange(len(input_cells), dtype=np.int32)
    del insert_position

    # Send sizes to create data structures for receiving from NeighAlltoAllv
    recv_size = np.zeros(len(source_ranks), dtype=np.int32)
    mesh_to_data_comm = comm.Create_dist_graph_adjacent(
        source_ranks.tolist(), dest_ranks.tolist(), reorder=False
    )
    mesh_to_data_comm.Neighbor_alltoall(dest_size, recv_size)

    # Prepare data-structures for receiving
    total_incoming = sum(recv_size)
    inc_cells = np.zeros(total_incoming, dtype=np.int64)
    inc_pos = np.zeros(total_incoming, dtype=np.intc)

    # Compute incoming offset
    inc_offsets = np.zeros(len(recv_size) + 1, dtype=np.intc)
    inc_offsets[1:] = np.cumsum(recv_size)

    # Send data
    s_msg = [out_cells, dest_size, MPI.INT64_T]
    r_msg = [inc_cells, recv_size, MPI.INT64_T]
    mesh_to_data_comm.Neighbor_alltoallv(s_msg, r_msg)

    s_msg = [out_pos, dest_size, MPI.INT32_T]
    r_msg = [inc_pos, recv_size, MPI.INT32_T]
    mesh_to_data_comm.Neighbor_alltoallv(s_msg, r_msg)
    mesh_to_data_comm.Free()

    local_input_range = compute_local_range(comm, num_cells_global)
    values_to_distribute = np.zeros_like(inc_pos, dtype=values.dtype)

    # Map values based on input cells and dofmap
    local_cells = inc_cells - local_input_range[0]
    values_to_distribute = values[dofmap_offsets[local_cells] + inc_pos]

    # Send input dofs back to owning process
    data_to_mesh_comm = comm.Create_dist_graph_adjacent(
        dest_ranks.tolist(), source_ranks.tolist(), reorder=False
    )

    incoming_global_dofs = np.zeros(sum(dest_size), dtype=values.dtype)
    s_msg = [values_to_distribute, recv_size, numpy_to_mpi[values.dtype.type]]
    r_msg = [incoming_global_dofs, dest_size, numpy_to_mpi[values.dtype.type]]
    data_to_mesh_comm.Neighbor_alltoallv(s_msg, r_msg)

    # Sort incoming global dofs as they were inputted
    assert len(incoming_global_dofs) == len(input_cells)
    sorted_global_dofs = np.zeros_like(incoming_global_dofs, dtype=values.dtype)
    sorted_global_dofs[proc_to_dof] = incoming_global_dofs

    data_to_mesh_comm.Free()
    return sorted_global_dofs


def send_and_recv_cell_perm(
    cells: npt.NDArray[np.int64],
    perms: npt.NDArray[np.uint32],
    cell_owners: npt.NDArray[np.int32],
    comm: MPI.Intracomm,
) -> tuple[npt.NDArray[np.int64], npt.NDArray[np.uint32]]:
    """
    Send global cell index and permutation to corresponding entry in `dest_ranks`.

    Args:
        cells: The global input index of the cell
        perms: The corresponding cell permutation of the cell
        cell_owners: The rank to send the i-th entry of cells and perms to
        comm: Rank of comm to generate neighbourhood communicator from
    """
    dest_ranks, _dest_size = np.unique(cell_owners, return_counts=True)
    dest_size = _dest_size.astype(np.int32)
    del _dest_size

    mesh_to_data = comm.Create_dist_graph(
        [comm.rank], [len(dest_ranks)], dest_ranks.tolist(), reorder=False
    )
    source, dest, _ = mesh_to_data.Get_dist_neighbors()
    assert np.allclose(dest, dest_ranks)
    insert_position = compute_insert_position(cell_owners, dest_ranks.astype(np.int32), dest_size)

    # Pack cells and permutations for sending
    out_cells = np.zeros_like(cells, dtype=np.int64)
    out_perm = np.zeros_like(perms, dtype=np.uint32)
    out_cells[insert_position] = cells
    out_perm[insert_position] = perms
    del insert_position

    # Send sizes to create data structures for receiving from NeighAlltoAllv
    recv_size = np.zeros_like(source, dtype=np.int32)
    mesh_to_data.Neighbor_alltoall(dest_size, recv_size)

    # Prepare data-structures for receiving
    total_incoming = sum(recv_size)
    inc_cells = np.zeros(total_incoming, dtype=np.int64)
    inc_perm = np.zeros(total_incoming, dtype=np.uint32)

    # Compute incoming offset
    inc_offsets = np.zeros(len(recv_size) + 1, dtype=np.intc)
    inc_offsets[1:] = np.cumsum(recv_size)

    # Send data
    s_msg = [out_cells, dest_size, MPI.INT64_T]
    r_msg = [inc_cells, recv_size, MPI.INT64_T]
    mesh_to_data.Neighbor_alltoallv(s_msg, r_msg)

    s_msg = [out_perm, dest_size, MPI.UINT32_T]
    r_msg = [inc_perm, recv_size, MPI.UINT32_T]
    mesh_to_data.Neighbor_alltoallv(s_msg, r_msg)
    mesh_to_data.Free()
    return inc_cells, inc_perm


def send_dofs_and_recv_values(
    input_dofmap: npt.NDArray[np.int64],
    dofmap_owners: npt.NDArray[np.int32],
    comm: MPI.Intracomm,
    input_array: npt.NDArray[valid_function_types],
    array_start: int,
):
    """
    Send a set of dofs (global index) to the process holding the DOF values to retrieve them.

    Args:
        input_dofmap: List of dofs (global index) that this process wants values for
        dofmap_owners: The process currently holding the values this process want to get.
        comm: MPI communicator
        input_array: Values for dofs
        array_start: The global starting index of `input_array`.
    """
    dest_ranks, _dest_size = np.unique(dofmap_owners, return_counts=True)
    dest_size = _dest_size.astype(np.int32)
    del _dest_size

    dofmap_to_values = comm.Create_dist_graph(
        [comm.rank], [len(dest_ranks)], dest_ranks.tolist(), reorder=False
    )

    source, dest, _ = dofmap_to_values.Get_dist_neighbors()
    assert np.allclose(dest_ranks, dest)
    # Compute amount of data to send to each process

    insert_position = compute_insert_position(dofmap_owners, dest_ranks, dest_size)

    # Pack dofs for sending
    out_dofs = np.zeros(len(dofmap_owners), dtype=np.int64)
    out_dofs[insert_position] = input_dofmap

    # Compute map from the data index sent to each process and the local number on
    # the current process
    proc_to_local = np.zeros_like(input_dofmap, dtype=np.int32)
    proc_to_local[insert_position] = np.arange(len(input_dofmap), dtype=np.int32)
    del insert_position

    # Send sizes to create data structures for receiving from NeighAlltoAllv
    recv_size = np.zeros_like(source, dtype=np.int32)
    recv_size.resize(max(len(recv_size), 1))  # Minimal resize to work with ompi
    dest_size.resize(max(len(dest_size), 1))  # Mininal resize to work with ompi
    dofmap_to_values.Neighbor_alltoall(dest_size, recv_size)
    dest_size.resize(len(dest))
    recv_size.resize(len(source))

    # Send input dofs to processes holding input array
    inc_dofs = np.zeros(sum(recv_size), dtype=np.int64)
    s_msg = [out_dofs, dest_size, MPI.INT64_T]
    r_msg = [inc_dofs, recv_size, MPI.INT64_T]
    dofmap_to_values.Neighbor_alltoallv(s_msg, r_msg)
    dofmap_to_values.Free()

    # Send back appropriate input values
    if len(input_array) > 0:
        sending_values = input_array[inc_dofs - array_start]
    else:
        sending_values = np.zeros(0, dtype=input_array.dtype)

    values_to_dofmap = comm.Create_dist_graph_adjacent(dest, source, reorder=False)
    inc_values = np.zeros_like(out_dofs, dtype=input_array.dtype)
    s_msg_rev = [sending_values, recv_size, numpy_to_mpi[input_array.dtype.type]]
    r_msg_rev = [inc_values, dest_size, numpy_to_mpi[input_array.dtype.type]]
    values_to_dofmap.Neighbor_alltoallv(s_msg_rev, r_msg_rev)
    values_to_dofmap.Free()

    # Sort inputs according to local dof number (input process)
    values = np.empty_like(inc_values, dtype=input_array.dtype)
    values[proc_to_local] = inc_values
    return values



"""
H5py interface to io4dolfinx

SPDX License identifier: MIT

Copyright: Jørgen S. Dokken, Henrik N.T. Finsberg, Simula Research Laboratory
"""

import contextlib
from pathlib import Path
from typing import Any

from mpi4py import MPI

import dolfinx
import h5py
import numpy as np
import numpy.typing as npt
from dolfinx.graph import adjacencylist

from ...structures import ArrayData, FunctionData, MeshData, MeshTagsData, ReadMeshData
from ...utils import check_file_exists, compute_local_range
from .. import FileMode, ReadMode

read_mode = ReadMode.parallel

# try:
# except ModuleNotFoundError:
#     raise ModuleNotFoundError("This backend requires h5py to be installed.")


@contextlib.contextmanager
def h5pyfile(h5name, filemode="r", force_serial: bool = False, comm=None):
    """Context manager for opening an HDF5 file with h5py.

    Args:
        h5name: The name of the HDF5 file.
        filemode: The file mode.
        force_serial: Force serial access to the file.
        comm: The MPI communicator

    """
    if comm is None:
        comm = MPI.COMM_WORLD

    if h5py.h5.get_config().mpi and not force_serial:
        h5file = h5py.File(h5name, filemode, driver="mpio", comm=comm)
    else:
        if comm.size > 1 and not force_serial:
            raise ValueError(
                f"h5py is not installed with MPI support, while using {comm.size} processes.",
                "If you really want to do this, turn on the `force_serial` flag.",
            )
        h5file = h5py.File(h5name, filemode)

    try:
        yield h5file
    finally:
        if h5file.id:
            h5file.close()


def get_default_backend_args(arguments: dict[str, Any] | None) -> dict[str, Any]:
    args = arguments or {"legacy": False}  # If meshtags is read from legacy
    return args


def convert_file_mode(mode: FileMode) -> str:
    match mode:
        case FileMode.append:
            return "a"
        case FileMode.read:
            return "r"
        case FileMode.write:
            return "w"
        case _:
            raise NotImplementedError(f"File mode {mode} not implemented")


def write_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    attributes: dict[str, np.ndarray],
    backend_args: dict[str, Any] | None = None,
):
    """Write attributes to file using H5PY.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        name: Name of the attributes
        attributes: Dictionary of attributes to write to file
        engine: ADIOS2 engine to use
    """
    filemode = "a" if Path(filename).exists() else "w"
    with h5pyfile(filename, filemode=filemode, comm=comm, force_serial=False) as h5file:
        if name not in h5file.keys():
            h5file.create_group(name, track_order=True)
        group = h5file[name]

        for key, val in attributes.items():
            group.attrs[key] = val


def read_attributes(
    filename: Path | str,
    comm: MPI.Intracomm,
    name: str,
    backend_args: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Read attributes from file using H5PY.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the attributes
    Returns:
        The attributes
    """
    check_file_exists(filename)
    output_attrs = {}
    with h5pyfile(filename, filemode="r", comm=comm, force_serial=False) as h5file:
        for key, val in h5file[name].attrs.items():
            output_attrs[key] = val
    return output_attrs


def read_timestamps(
    filename: Path | str,
    comm: MPI.Intracomm,
    function_name: str,
    backend_args: dict[str, Any] | None = None,
) -> npt.NDArray[np.float64 | str]:  # type: ignore[type-var]
    """Read time-stamps from a checkpoint file.

    Args:
        comm: MPI communicator
        filename: Path to file
        function_name: Name of the function to read time-stamps for
        backend_args: Arguments for backend, for instance file type.

    Returns:
        The time-stamps
    """
    check_file_exists(filename)
    mesh_name = "mesh"
    with h5pyfile(filename, filemode="r", comm=comm, force_serial=False) as h5file:
        mesh_directory = h5file[mesh_name]
        functions = mesh_directory["functions"]
        u = functions[function_name]
        timestamps = u.attrs["timestamps"]
    return timestamps


def write_mesh(
    filename: Path | str,
    comm: MPI.Intracomm,
    mesh: MeshData,
    backend_args: dict[str, Any] | None = None,
    mode: FileMode = FileMode.write,
    time: float = 0.0,
):
    """Write a mesh to file using H5PY

    Args:
        comm: MPI communicator used in storage
        mesh: Internal data structure for the mesh data to save to file
        filename: Path to file to write to.
        mode: Mode to use (write or append)
        time: Time stamp
    """
    backend_args = get_default_backend_args(backend_args)
    h5_mode = convert_file_mode(mode)
    mesh_name = "mesh"
    with h5pyfile(filename, filemode=h5_mode, comm=comm, force_serial=False) as h5file:
        if mesh_name in h5file.keys() and h5_mode == "a":
            mesh_directory = h5file[mesh_name]
            timestamps = mesh_directory.attrs["timestamps"]
            if np.isclose(time, timestamps).any():
                raise ValueError("Mesh has already been stored at time={time_stamp}.")
            else:
                mesh_directory.attrs["timestamps"] = np.append(
                    mesh_directory.attrs["timestamps"], time
                )
                idx = len(mesh_directory.attrs["timestamps"]) - 1
                write_topology = False
        else:
            mesh_directory = h5file.create_group(mesh_name)
            mesh_directory.attrs["timestamps"] = np.array([time], dtype=np.float64)
            idx = 0
            write_topology = True

        geometry_group = mesh_directory.create_group(f"{idx}")

        # Write geometry data
        gdim = mesh.local_geometry.shape[1]
        geometry_dataset = geometry_group.create_dataset(
            "Points",
            [mesh.num_nodes_global, gdim],
            dtype=mesh.local_geometry.dtype,
        )
        geometry_dataset[slice(*mesh.local_geometry_pos), :] = mesh.local_geometry

        # Write static partitioning data
        if "PartitioningData" not in mesh_directory.keys() and mesh.store_partition:
            assert mesh.partition_range is not None
            assert mesh.ownership_array is not None
            par_dataset = mesh_directory.create_dataset(
                "PartitioningData", [mesh.partition_global], dtype=mesh.ownership_array.dtype
            )
            par_dataset[slice(*mesh.partition_range)] = mesh.ownership_array

        if "PartitioningOffset" not in mesh_directory.keys() and mesh.store_partition:
            assert mesh.ownership_offset is not None
            par_dataset = mesh_directory.create_dataset(
                "PartitioningOffset", [mesh.num_cells_global + 1], dtype=np.int64
            )
            par_dataset[mesh.local_topology_pos[0] : mesh.local_topology_pos[1] + 1] = (
                mesh.ownership_offset
            )

        if "PartitionProcesses" not in mesh_directory.attrs.keys() and mesh.store_partition:
            mesh_directory.attrs["PartitionProcesses"] = mesh.partition_processes

        # Write static data
        if write_topology:
            mesh_directory.attrs["CellType"] = mesh.cell_type
            mesh_directory.attrs["Degree"] = mesh.degree
            mesh_directory.attrs["LagrangeVariant"] = mesh.lagrange_variant
            num_dofs_per_cell = mesh.local_topology.shape[1]
            topology_dataset = mesh_directory.create_dataset(
                "Topology", [mesh.num_cells_global, num_dofs_per_cell], dtype=np.int64
            )
            topology_dataset[slice(*mesh.local_topology_pos), :] = mesh.local_topology


def read_mesh_data(
    filename: Path | str,
    comm: MPI.Intracomm,
    time: str | float | None = 0.0,
    read_from_partition: bool = False,
    backend_args: dict[str, Any] | None = None,
) -> ReadMeshData:
    """Read mesh data from h5py based checkpoint files.

    Args:
        filename: Path to input file
        comm: The MPI communciator to distribute the mesh over
        time: Time stamp associated with mesh
        read_from_partition: Read mesh with partition from file
    Returns:
        The mesh topology, geometry, UFL domain and partition function
    """

    backend_args = get_default_backend_args(backend_args)

    with h5pyfile(filename, filemode="r", comm=comm, force_serial=False) as h5file:
        if "mesh" not in h5file.keys():
            raise KeyError("Could not find mesh in file")
        mesh_group = h5file["mesh"]
        timestamps = mesh_group.attrs["timestamps"]
        assert time is not None
        parent_group = np.flatnonzero(np.isclose(timestamps, time))
        if len(parent_group) != 1:
            raise RuntimeError(
                f"Time step {time} not found in file, available steps is {timestamps}"
            )
        time_group = f"{parent_group[0]:d}"

        # Get mesh topology (distributed)
        topology = mesh_group["Topology"]
        local_range = compute_local_range(comm, topology.shape[0])
        mesh_topology = topology[slice(*local_range), :]

        cell_type = mesh_group.attrs["CellType"]
        lvar = mesh_group.attrs["LagrangeVariant"]
        degree = mesh_group.attrs["Degree"]

        geometry_group = mesh_group[time_group]
        geometry_dataset = geometry_group["Points"]
        x_shape = geometry_dataset.shape

        geometry_range = compute_local_range(comm, x_shape[0])
        mesh_geometry = geometry_dataset[slice(*geometry_range), :]

        # Check validity of partitioning information
        if read_from_partition:
            if "PartitionProcesses" not in mesh_group.attrs.keys():
                raise KeyError(f"Partitioning information not found in {filename}")
            par_keys = ("PartitioningData", "PartitioningOffset")
            for key in par_keys:
                if key not in mesh_group.keys():
                    raise KeyError(f"Partitioning information not found in {filename}")

            par_num_procs = mesh_group.attrs["PartitionProcesses"]
            if par_num_procs != comm.size:
                raise ValueError(f"Number of processes in file ({par_num_procs})!=({comm.size=})")

            # First read in offsets based on the number of cells [0, num_cells_local]
            par_offsets = mesh_group["PartitioningOffset"][local_range[0] : local_range[1] + 1]
            # Then read the data based of offsets
            par_data = mesh_group["PartitioningData"][par_offsets[0] : par_offsets[-1]]
            # Then make offsets local
            par_offsets[:] -= par_offsets[0]
            partition_graph = adjacencylist(par_data, par_offsets.astype(np.int32))
        else:
            partition_graph = None

    return ReadMeshData(
        cells=mesh_topology,
        cell_type=cell_type,
        x=mesh_geometry,
        degree=degree,
        lvar=lvar,
        partition_graph=partition_graph,
    )


def write_meshtags(
    filename: str | Path,
    comm: MPI.Intracomm,
    data: MeshTagsData,
    backend_args: dict[str, Any] | None = None,
):
    """Write mesh tags to file.

    Args:
        filename: Path to file to write to
        comm: MPI communicator used in storage
        data: Internal data structure for the mesh tags to save to file
        backend_args: Arguments to backend
    """
    backend_args = get_default_backend_args(backend_args)

    with h5pyfile(filename, filemode="a", comm=comm, force_serial=False) as h5file:
        if "mesh" not in h5file.keys():
            raise KeyError("Could not find mesh in file")
        mesh_group = h5file["mesh"]
        if "tags" not in mesh_group.keys():
            tags = mesh_group.create_group("tags")
        else:
            tags = mesh_group["tags"]
        if data.name in tags.keys():
            raise KeyError(f"MeshTags with {data.name=} already exists in this file")
        tag = tags.create_group(data.name)

        # Add topology
        topology = tag.create_dataset(
            "Topology", shape=[data.num_entities_global, data.num_dofs_per_entity], dtype=np.int64
        )
        assert data.local_start is not None
        topology[data.local_start : data.local_start + len(data.indices), :] = data.indices

        # Add cell_type attribute
        tag.attrs["CellType"] = data.cell_type

        # Add values
        values = tag.create_dataset(
            "Values", shape=[data.num_entities_global], dtype=data.values.dtype
        )
        values[data.local_start : data.local_start + len(data.indices)] = data.values

        # Add dimension
        tag.attrs["dim"] = data.dim


def read_meshtags_data(
    filename: str | Path, comm: MPI.Intracomm, name: str, backend_args: dict[str, Any] | None = None
) -> MeshTagsData:
    """Read mesh tags from file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the mesh tags to read
        backend_args: Arguments to backend. If "legacy_dolfin" is supplied as argument
            the HDF5 file is assumed to have been made with DOLFIN

    Returns:
        Internal data structure for the mesh tags read from file
    """
    backend_args = get_default_backend_args(backend_args)
    legacy = backend_args["legacy"]
    with h5pyfile(filename, filemode="r", comm=comm, force_serial=False) as h5file:
        if legacy:
            if name not in h5file.keys():
                raise RuntimeError(f"MeshTag {name} not found in {filename}.")
            mesh = h5file[name]
            topology = mesh["topology"]
            cell_type = topology.attrs["celltype"]
            if isinstance(cell_type, np.bytes_):
                cell_type = cell_type.decode("utf-8")
            dim = dolfinx.mesh.cell_dim(dolfinx.mesh.to_type(cell_type))
            values = mesh["values"]
        else:
            if "mesh" not in h5file.keys():
                raise KeyError("No mesh found")
            mesh = h5file["mesh"]
            if "tags" not in mesh.keys():
                raise KeyError("Could not find 'tags' in file, are you sure this is a checkpoint?")
            tags = mesh["tags"]
            if name not in tags.keys():
                raise KeyError(f"Could not find {name} in '/mesh/tags/' in {filename}")
            tag = tags[name]

            dim = tag.attrs["dim"]
            topology = tag["Topology"]
            values = tag["Values"]
        num_entities_global = topology.shape[0]
        topology_range = compute_local_range(comm, num_entities_global)
        indices = topology[slice(*topology_range), :].astype(np.int64)
        vals = values[slice(*topology_range)].astype(np.int32)
        return MeshTagsData(name=name, values=vals, indices=indices, dim=dim)


def read_dofmap(
    filename: str | Path, comm: MPI.Intracomm, name: str, backend_args: dict[str, Any] | None
) -> dolfinx.graph.AdjacencyList:
    """Read the dofmap of a function with a given name.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofmap for
        backend_args: Arguments to backend

    Returns:
        Dofmap as an {py:class}`dolfinx.graph.AdjacencyList`
    """
    backend_args = {} if backend_args is None else backend_args
    with h5pyfile(filename, filemode="r", comm=comm, force_serial=False) as h5file:
        # If dofmap is read with full path, it is passed through backend_args
        dofmap_key = backend_args.get("dofmap", None)
        if dofmap_key is None:
            mesh_name = "mesh"  # Prepare for multiple meshes
            if mesh_name not in h5file.keys():
                raise KeyError(f"No mesh '{mesh_name}' found in {filename}")
            mesh = h5file[mesh_name]
            if "functions" not in mesh.keys():
                raise KeyError(f"No functions stored in '{mesh_name}' in {filename}")
            functions = mesh["functions"]
            if name not in functions.keys():
                raise KeyError(
                    f"No function with name '{name}' on '{mesh_name}' stored in {filename}"
                )
            function = functions[name]
            offset_key = "dofmap_offsets"
            dofmap_key = "dofmap"
            offsets = function[offset_key]
            dofmap = function[dofmap_key]
        else:
            offset_key = backend_args["offsets"]
            dofmap = h5file[dofmap_key]
            offsets = h5file[offset_key]

        num_cells = offsets.shape[0] - 1
        local_range = compute_local_range(comm, num_cells)

        # First read in offsets based on the number of cells [0, num_cells_local]
        glob_offsets = offsets[local_range[0] : local_range[1] + 1].flatten().astype(np.int64)

        # Then read the data based of offsets
        dofmap_data = dofmap[glob_offsets[0] : glob_offsets[-1]].flatten()

    # Then make offsets local
    loc_offsets = (glob_offsets - glob_offsets[0]).astype(np.int32)
    return adjacencylist(dofmap_data, loc_offsets)


def read_dofs(
    filename: str | Path,
    comm: MPI.Intracomm,
    name: str,
    time: float,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.float32 | np.float64 | np.complex64 | np.complex128], int]:
    """Read the dofs (values) of a function with a given name from a given timestep.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        name: Name of the function to read the dofs for
        time: Time stamp associated with the function to read
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of degrees of freedom (with respect to input data)
        and the global starting point on the process.
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    with h5pyfile(filename, filemode="r", comm=comm, force_serial=False) as h5file:
        mesh_name = "mesh"  # Prepare for multiple meshes
        if mesh_name not in h5file.keys():
            raise RuntimeError(f"No mesh '{mesh_name}' found in {filename}")
        mesh = h5file[mesh_name]
        if "functions" not in mesh.keys():
            raise RuntimeError(f"No functions stored in '{mesh_name}' in {filename}")
        functions = mesh["functions"]
        if name not in functions.keys():
            raise RuntimeError(
                f"No function with name '{name}' on '{mesh_name}' stored in {filename}"
            )
        function = functions[name]
        timestamps = function.attrs["timestamps"]
        idx = np.flatnonzero(np.isclose(timestamps, time))
        if len(idx) != 1:
            raise RuntimeError("Could not find {name}(t={time}) on grid {mesh_name} in {filename}.")
        u_t = function[f"{idx[0]:d}"]
        data_group = u_t["array"]
        num_dofs_global = data_group.shape[0]
        local_range = compute_local_range(comm, num_dofs_global)
        local_array = data_group[slice(*local_range)]
        return local_array, local_range[0]


def read_cell_perms(
    comm: MPI.Intracomm, filename: Path | str, backend_args: dict[str, Any] | None
) -> npt.NDArray[np.uint32]:
    """
    Read cell permutation from file with given communicator,
    Split in continuous chunks based on number of cells in the input data.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        backend_args: Arguments to backend

    Returns:
        Contiguous sequence of permutations (with respect to input data)
        Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """

    with h5pyfile(filename, filemode="r", comm=comm, force_serial=False) as h5file:
        mesh_name = "mesh"  # Prepare for multiple meshes
        if mesh_name not in h5file.keys():
            raise RuntimeError(f"No mesh '{mesh_name}' found in {filename}")
        mesh = h5file[mesh_name]
        data_group = mesh["CellPermutations"]
        num_cells_global = data_group.shape[0]
        local_range = compute_local_range(comm, num_cells_global)
        local_array = data_group[slice(*local_range)]
        return local_array


def write_function(
    filename: str | Path,
    comm: MPI.Intracomm,
    u: FunctionData,
    time: float,
    mode: FileMode,
    backend_args: dict[str, Any] | None = None,
):
    """Write a function to file.

    Args:
        comm: MPI communicator used in storage
        u: Internal data structure for the function data to save to file
        filename: Path to file to write to
        time: Time stamp associated with function
        mode: File-mode to store the function
        backend_args: Arguments to backend
    """

    mesh_name = "mesh"  # Prepare for multiple meshes
    backend_args = get_default_backend_args(backend_args)
    h5_mode = convert_file_mode(mode)
    with h5pyfile(filename, filemode=h5_mode, comm=comm, force_serial=False) as h5file:
        cell_permutations_exist = False
        dofmap_exists = False
        dofmap_offsets_exists = False
        if h5_mode == "a":
            if mesh_name not in h5file.keys():
                mesh = h5file.create_group(mesh_name)
            else:
                mesh = h5file[mesh_name]

            cell_permutations_exist = "CellPermutations" in mesh.keys()

            if "functions" not in mesh.keys():
                functions = mesh.create_group("functions")
            else:
                functions = mesh["functions"]

            if u.name not in functions.keys():
                function = functions.create_group(u.name)
            else:
                function = functions[u.name]

            dofmap_exists = "dofmap" in function.keys()
            dofmap_offsets_exists = "dofmap_offsets" in function.keys()

        if not cell_permutations_exist:
            cell_perms = mesh.create_dataset(
                "CellPermutations", shape=[u.num_cells_global], dtype=np.uint32
            )
            cell_perms[slice(*u.local_cell_range)] = u.cell_permutations

        if not dofmap_exists:
            dofmap = function.create_dataset(
                "dofmap", shape=[u.global_dofs_in_dofmap], dtype=np.int64
            )
            dofmap[slice(*u.dofmap_range)] = u.dofmap_array

        if not dofmap_offsets_exists:
            dofmap_offsets = function.create_dataset(
                "dofmap_offsets", shape=[u.num_cells_global + 1], dtype=np.int64
            )
            dofmap_offsets[u.local_cell_range[0] : u.local_cell_range[1] + 1] = u.dofmap_offsets

        # Write timestamp
        if "timestamps" in function.attrs.keys():
            timestamps = function.attrs["timestamps"]
            if np.isclose(time, timestamps).any():
                raise RuntimeError("FUnction has already been stored at time={time_stamp}.")
            else:
                function.attrs["timestamps"] = np.append(function.attrs["timestamps"], time)
        else:
            function.attrs["timestamps"] = np.array([time])
        idx = len(function.attrs["timestamps"]) - 1

        data_group = function.create_group(f"{idx:d}")
        array = data_group.create_dataset("array", shape=[u.num_dofs_global], dtype=u.values.dtype)
        array[slice(*u.dof_range)] = u.values


def read_legacy_mesh(
    filename: Path | str, comm: MPI.Intracomm, group: str
) -> tuple[npt.NDArray[np.int64], npt.NDArray[np.floating], str | None]:
    """Read in the mesh topology, geometry and (optionally) cell type from a
    legacy DOLFIN HDF5-file.

    Args:
        filename: Path to file to read from
        comm: MPI communicator used in storage
        group: Group in HDF5 file where mesh is stored

    Returns:
        Tuple containing:
            - Topology as a (num_cells, num_vertices_per_cell) array of global vertex indices
            - Geometry as a (num_vertices, geometric_dimension) array of vertex coordinates
            - Cell type as a string (e.g. "tetrahedron") or None if not found
    """
    with h5pyfile(filename, filemode="r", comm=comm, force_serial=False) as h5file:
        if group not in h5file.keys():
            raise KeyError(f"Could not find {group}  in {filename}.")
        mesh = h5file[group]
        if "topology" not in mesh.keys():
            raise KeyError(f"Could not find '{group}/topology'  in {filename}.")

        topology = mesh["topology"]
        shape = topology.shape
        local_range = compute_local_range(comm, shape[0])
        mesh_topology = topology[slice(*local_range)].astype(np.int64)

        # Get mesh cell type
        cell_type = None
        if "celltype" in topology.attrs.keys():
            cell_type = topology.attrs["celltype"]
            if isinstance(cell_type, np.bytes_):
                cell_type = cell_type.decode("utf-8")

        for geometry_key in ["geometry", "coordinates"]:
            if geometry_key in mesh.keys():
                break
        else:
            raise KeyError(
                "Could not find geometry in '/mesh/geometry' or '/mesh/coordinates'"
                + f" in {filename}."
            )
        geometry = mesh[geometry_key]
        g_shape = geometry.shape
        local_g_range = compute_local_range(comm, g_shape[0])
        mesh_geometry = geometry[slice(*local_g_range)]

    return mesh_topology, mesh_geometry, cell_type


def read_hdf5_array(
    comm: MPI.Intracomm,
    filename: Path | str,
    group: str,
    backend_args: dict[str, Any] | None = None,
) -> tuple[np.ndarray, int]:
    """Read an array from an HDF5 file.

    Args:
        comm: MPI communicator used in storage
        filename: Path to file to read from
        group: Group in HDF5 file where array is stored
        backend_args: Arguments to backend

    Returns:
        Tuple containing:
            - Numpy array read from file
            - Global starting point on the process.
                Process 0 has [0, M), process 1 [M, N), process 2 [N, O) etc.
    """
    with h5pyfile(filename, filemode="r", comm=comm, force_serial=False) as h5file:
        data = h5file[group]
        shape = data.shape[0]
        local_range = compute_local_range(comm, shape)
        out_data = data[slice(*local_range)].flatten()
    return out_data, local_range[0]


def snapshot_checkpoint(
    filename: Path | str,
    mode: FileMode,
    u: dolfinx.fem.Function,
    backend_args: dict[str, Any] | None,
):
    """Create a snapshot checkpoint of a dolfinx function.

    Args:
        filename: Path to file to read from
        mode: File-mode to store the function
        u: dolfinx function to create a snapshot checkpoint for
        backend_args: Arguments to backend
    """
    comm = u.function_space.mesh.comm
    dofmap = u.function_space.dofmap
    local_range = np.array(dofmap.index_map.local_range) * dofmap.index_map_bs
    num_dofs_local = local_range[1] - local_range[0]
    num_dofs_global = dofmap.index_map.size_global * dofmap.index_map_bs
    h5mode = convert_file_mode(mode)
    if h5mode == "w":
        with h5pyfile(filename, filemode=h5mode, comm=comm, force_serial=False) as h5file:
            local_dofs = u.x.array[:num_dofs_local].copy()
            data = h5file.create_group("snapshot")
            dataset = data.create_dataset("dofs", shape=num_dofs_global, dtype=local_dofs.dtype)
            dataset[slice(*local_range)] = local_dofs
    elif h5mode == "r":
        with h5pyfile(filename, filemode=h5mode, comm=comm, force_serial=False) as h5file:
            data = h5file["snapshot"]["dofs"]
            assert data.shape[0] == num_dofs_global
            u.x.array[:num_dofs_local] = data[slice(*local_range)]
            u.x.scatter_forward()


def read_function_names(
    filename: Path | str, comm: MPI.Intracomm, backend_args: dict[str, Any] | None
) -> list[str]:
    """Read all function names from a file.

    Args:
        filename: Path to file
        comm: MPI communicator to launch IO on.
        backend_args: Arguments to backend

    Returns:
        A list of function names.
    """
    check_file_exists(filename)
    backend_args = get_default_backend_args(backend_args)
    with h5pyfile(filename, filemode="r", comm=comm, force_serial=False) as h5file:
        mesh_name = "mesh"  # Prepare for multiple meshes
        if mesh_name not in h5file.keys():
            raise RuntimeError(f"No mesh '{mesh_name}' found in {filename}")
        mesh = h5file[mesh_name]
        if "functions" not in mesh.keys():
            return []
        else:
            return list(mesh["functions"].keys())


def read_point_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: float | str | None,
    backend_args: dict[str, Any] | None,
) -> tuple[np.ndarray, int]:
    """Read data from the nodes of a mesh.

    Args:
        filename: Path to file
        name: Name of point data
        comm: Communicator to launch IO on.
        time: The time stamp
        backend_args: The backend arguments

    Returns:
       Data local to process (contiguous, no mpi comm) and local start range
    """
    raise NotImplementedError("The h5py backend cannot read point data.")


def read_cell_data(
    filename: Path | str,
    name: str,
    comm: MPI.Intracomm,
    time: str | float | None,
    backend_args: dict[str, Any] | None,
) -> tuple[npt.NDArray[np.int64], np.ndarray]:
    """Read data from the cells of a mesh.

    Args:
        filename: Path to file
        name: Name of point data
        comm: Communicator to launch IO on.
        time: The time stamp
        backend_args: The backend arguments
    Returns:
        A tuple (topology, dofs) where topology contains the
        vertex indices of the cells, dofs the degrees of
        freedom within that cell.
    """
    raise NotImplementedError("The h5py backend does not support reading cell data.")


def write_data(
    filename: Path | str,
    array_data: ArrayData,
    comm: MPI.Intracomm,
    time: str | float | None,
    mode: FileMode,
    backend_args: dict[str, Any] | None,
):
    """Write a 2D-array to file (distributed across proceses with MPI).

    Args:
        filename: Path to file
        array_data: Data to write to file
        comm: MPI communicator to open the file with
        time: Time-stamp for data.
        mode: Append or write
        backend_args: The backend arguments
    """
    raise NotImplementedError("H5PY has not implemented this yet")



# # Writing MeshTags data to a checkpoint file
# In many scenarios, the mesh used in a checkpoint is not trivial, and subdomains and sub-entities
# have been tagged with appropriate markers.
# As the mesh gets redistributed when read
# (see [Writing Mesh Checkpoint](./writing_mesh_checkpoint)),
# we need to store any tags together with this new mesh.

# As an example we will use a unit-cube, where each entity has been tagged with a unique index.

# +
import logging
from pathlib import Path

from mpi4py import MPI

import dolfinx
import ipyparallel as ipp
import numpy as np

import io4dolfinx

assert MPI.COMM_WORLD.size == 1, "This example should only be run with 1 MPI process"

mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, nx=3, ny=4, nz=5)
# -

# We start by computing the unique global index of each (owned) entity in the mesh
# as well as its corresponding midpoint

entity_midpoints = {}
meshtags = {}
for i in range(mesh.topology.dim + 1):
    mesh.topology.create_entities(i)
    e_map = mesh.topology.index_map(i)

    # Compute midpoints of entities
    entities = np.arange(e_map.size_local, dtype=np.int32)
    mesh.topology.create_connectivity(i, mesh.topology.dim)
    entity_midpoints[i] = dolfinx.mesh.compute_midpoints(mesh, i, entities)
    # Associate each local index with its global index
    values = np.arange(e_map.size_local, dtype=np.int32) + e_map.local_range[0]
    meshtags[i] = dolfinx.mesh.meshtags(mesh, i, entities, values)

# We use {py:func}`io4dolfinx.write_mesh` and `io4dolfinx.write_meshtags` to write the
# {py:class}`dolfinx.mesh.Mesh` and {py:class}`dolfinx.mesh.MeshTags` to file.
# We associate each meshtag with a name

filename = Path("mesh_with_meshtags.bp")
io4dolfinx.write_mesh(filename, mesh)
for i, tag in meshtags.items():
    io4dolfinx.write_meshtags(filename, mesh, tag, meshtag_name=f"meshtags_{i}")


# Next we want to read the meshtags in on a different number of processes,
# and check that the midpoints of each entity is still correct.
# We do this with {py:func}`io4dolfinx.read_meshtags`.


def verify_meshtags(filename: Path):
    # We assume that entity_midpoints have been sent to the engine
    from mpi4py import MPI

    import dolfinx
    import numpy as np

    import io4dolfinx

    read_mesh = io4dolfinx.read_mesh(filename, MPI.COMM_WORLD)
    prefix = f"{read_mesh.comm.rank + 1}/{read_mesh.comm.size}: "
    for i in range(read_mesh.topology.dim + 1):
        # Read mesh from file
        meshtags = io4dolfinx.read_meshtags(filename, read_mesh, meshtag_name=f"meshtags_{i}")

        # Compute midpoints for all local entities on process
        read_mesh.topology.create_connectivity(i, read_mesh.topology.dim)
        midpoints = dolfinx.mesh.compute_midpoints(read_mesh, i, meshtags.indices)
        # Compare locally computed midpoint with reference data
        for global_pos, midpoint in zip(meshtags.values, midpoints):
            np.testing.assert_allclose(
                entity_midpoints[i][global_pos],
                midpoint,
                err_msg=f"{prefix}: Midpoint ({i, global_pos}) do not match",
            )
        print(f"{prefix} Matching of all entities of dimension {i} successful")


with ipp.Cluster(engines="mpi", n=3, log_level=logging.ERROR) as cluster:
    cluster[:].push({"entity_midpoints": entity_midpoints})
    query = cluster[:].apply_async(verify_meshtags, filename)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))



# Copyright (C) 2023 Jørgen Schartum Dokken
#
# This file is part of io4dolfinx
#
# SPDX-License-Identifier:    MIT

"""
Vectorized numpy operations used internally in io4dolfinx
"""

from __future__ import annotations

from pathlib import Path

from mpi4py import MPI

import basix.ufl
import dolfinx
import numpy as np
import numpy.typing as npt
import ufl
from packaging.version import Version

__all__ = [
    "check_file_exists",
    "compute_local_range",
    "index_owner",
    "compute_dofmap_pos",
    "unroll_dofmap",
    "compute_insert_position",
    "unroll_insert_position",
    "reconstruct_mesh",
]


def check_file_exists(filename: Path | str):
    """Check if file exists."""
    if not Path(filename).exists():
        raise FileNotFoundError(f"{filename} not found")


valid_function_types = np.float32 | np.float64 | np.complex64 | np.complex128
valid_real_types = np.float32 | np.float64


def element_signature(V):
    if Version(dolfinx.__version__) > Version("0.9.0"):
        return V.element.signature
    else:
        return V.element.signature()


def compute_insert_position(
    data_owner: npt.NDArray[np.int32],
    destination_ranks: npt.NDArray[np.int32],
    out_size: npt.NDArray[np.int32],
) -> npt.NDArray[np.int32]:
    """
    Giving a list of ranks, compute the local insert position for each rank in a list
    sorted by destination ranks. This function is used for packing data from a
    given process to its destination processes.

    Example:

        .. highlight:: python
        .. code-block:: python

            data_owner = [0, 1, 1, 0, 2, 3]
            destination_ranks = [2,0,3,1]
            out_size = [1, 2, 1, 2]
            insert_position = compute_insert_position(data_owner, destination_ranks, out_size)

        Insert position is then ``[1, 4, 5, 2, 0, 3]``
    """
    process_pos_indicator = data_owner.reshape(-1, 1) == destination_ranks

    # Compute offsets for insertion based on input size
    send_offsets = np.zeros(len(out_size) + 1, dtype=np.intc)
    send_offsets[1:] = np.cumsum(out_size)
    assert send_offsets[-1] == len(data_owner)

    # Compute local insert index on each process
    proc_row, proc_col = np.nonzero(process_pos_indicator)
    cum_pos = np.cumsum(process_pos_indicator, axis=0)
    insert_position = cum_pos[proc_row, proc_col] - 1

    # Add process offset for each local index
    insert_position += send_offsets[proc_col]
    return insert_position


def unroll_insert_position(
    insert_position: npt.NDArray[np.int32], block_size: int
) -> npt.NDArray[np.int32]:
    """
    Unroll insert position by a block size

    Example:


        .. highlight:: python
        .. code-block:: python

            insert_position = [1, 4, 5, 2, 0, 3]
            unrolled_ip = unroll_insert_position(insert_position, 3)

        where ``unrolled_ip = [3, 4 ,5, 12, 13, 14, 15, 16, 17, 6, 7, 8, 0, 1, 2, 9, 10, 11]``
    """
    unrolled_ip = np.repeat(insert_position, block_size) * block_size
    unrolled_ip += np.tile(np.arange(block_size), len(insert_position))
    return unrolled_ip


def compute_local_range(comm: MPI.Intracomm, N: np.int64):
    """
    Divide a set of `N` objects into `M` partitions, where `M` is
    the size of the MPI communicator `comm`.

    NOTE: If N is not divisible by the number of ranks, the first `r`
    processes gets an extra value

    Returns the local range of values
    """
    rank = comm.rank
    size = comm.size
    n = N // size
    r = N % size
    # First r processes has one extra value
    if rank < r:
        return [rank * (n + 1), (rank + 1) * (n + 1)]
    else:
        return [rank * n + r, (rank + 1) * n + r]


def index_owner(
    comm: MPI.Intracomm, indices: npt.NDArray[np.int64], N: np.int64
) -> npt.NDArray[np.int32]:
    """
    Find which rank (local to comm) which owns an `index`, given that
    data of size `N` has been split equally among the ranks.

    NOTE: If `N` is not divisible by the number of ranks, the first `r`
    processes gets an extra value.
    """
    size = comm.size
    assert (indices < N).all()
    n = N // size
    r = N % size

    owner = np.empty_like(indices, dtype=np.int32)
    inc_remainder = indices < (n + 1) * r
    owner[inc_remainder] = indices[inc_remainder] // (n + 1)
    owner[~inc_remainder] = r + (indices[~inc_remainder] - r * (n + 1)) // n
    return owner


def unroll_dofmap(dofs: npt.NDArray[np.int32], bs: int) -> npt.NDArray[np.int32]:
    """
    Given a two-dimensional dofmap of size `(num_cells, num_dofs_per_cell)`
    Expand the dofmap by its block size such that the resulting array
    is of size `(num_cells, bs*num_dofs_per_cell)`
    """
    num_cells, num_dofs_per_cell = dofs.shape
    unrolled_dofmap = np.repeat(dofs, bs).reshape(num_cells, num_dofs_per_cell * bs) * bs
    unrolled_dofmap += np.tile(np.arange(bs), num_dofs_per_cell)
    return unrolled_dofmap


def compute_dofmap_pos(
    V: dolfinx.fem.FunctionSpace,
) -> tuple[npt.NDArray[np.int32], npt.NDArray[np.int32]]:
    """
    Compute a map from each owned dof in the dofmap to a single cell owned by the
    process, and the relative position of the dof.

    :param V: The function space
    :returns: The tuple (`cells`, `dof_pos`) where each array is the size of the
        number of owned dofs (unrolled for block size)
    """
    dofs = V.dofmap.list
    mesh = V.mesh
    num_owned_cells = mesh.topology.index_map(mesh.topology.dim).size_local
    dofmap_bs = V.dofmap.bs
    num_owned_dofs = V.dofmap.index_map.size_local * V.dofmap.index_map_bs

    local_cell = np.empty(
        num_owned_dofs, dtype=np.int32
    )  # Local cell index for each dof owned by process
    dof_pos = np.empty(num_owned_dofs, dtype=np.int32)  # Position in dofmap for said dof

    unrolled_dofmap = unroll_dofmap(dofs[:num_owned_cells, :], dofmap_bs)
    markers = unrolled_dofmap < num_owned_dofs
    local_indices = np.broadcast_to(np.arange(markers.shape[1]), markers.shape)
    cell_indicator = np.broadcast_to(
        np.arange(num_owned_cells, dtype=np.int32).reshape(-1, 1),
        (num_owned_cells, markers.shape[1]),
    )
    indicator = unrolled_dofmap[markers].reshape(-1)
    local_cell[indicator] = cell_indicator[markers].reshape(-1)
    dof_pos[indicator] = local_indices[markers].reshape(-1)
    return local_cell, dof_pos


def reconstruct_mesh(mesh: dolfinx.mesh.Mesh, coordinate_element_degree: int) -> dolfinx.mesh.Mesh:
    """
    Make a copy of a mesh and potentially change the element of the coordinate element.

    Note:
        The topology is shared with the original mesh but the geometry is reconstructed.

    Args:
        mesh: Mesh to reconstruct
        coordinate_element_degree: Degree to use for coordinate element

    Returns:
        The new mesh

    """
    # Extract cell properties
    ud = mesh.ufl_domain()
    assert ud is not None
    c_el = ud.ufl_coordinate_element()
    family = c_el.family_name
    lvar = c_el.lagrange_variant
    ct = c_el.cell_type

    # Create new UFL element
    new_c_el = basix.ufl.element(
        family,
        ct,
        coordinate_element_degree,
        shape=(mesh.geometry.dim,),
        lagrange_variant=lvar,
        dtype=mesh.geometry.x.dtype,
    )

    # Extract new node coordinates
    V_tmp = dolfinx.fem.functionspace(mesh, new_c_el)
    gdim = mesh.geometry.dim
    x = V_tmp.tabulate_dof_coordinates()[:, :gdim]

    # Create new geoemtry
    geom_imap = V_tmp.dofmap.index_map
    geom_dofmap = V_tmp.dofmap.list
    num_nodes_local = geom_imap.size_local + geom_imap.num_ghosts
    original_input_indices = geom_imap.local_to_global(np.arange(num_nodes_local, dtype=np.int32))
    coordinate_element = dolfinx.fem.coordinate_element(
        mesh.topology.cell_type, coordinate_element_degree, lvar, dtype=mesh.geometry.x.dtype
    )
    # Could use create_geometry here when things are fixed
    geom = dolfinx.mesh.Geometry(
        type(mesh.geometry._cpp_object)(
            geom_imap, geom_dofmap, coordinate_element._cpp_object, x, original_input_indices
        )
    )

    # Create new mesh
    new_top = mesh.topology
    cpp_mesh = type(mesh._cpp_object)(mesh.comm, new_top._cpp_object, geom._cpp_object)
    return dolfinx.mesh.Mesh(cpp_mesh, ufl.Mesh(new_c_el))



# Copyright (C) 2023 Jørgen Schartum Dokken
#
# This file is part of io4dolfinx
#
# SPDX-License-Identifier:    MIT


"""
Functions to create checkpoints with Legacy dolfin
"""

import argparse
import pathlib

import dolfin
import numpy as np
import ufl_legacy as ufl


def create_reference_data(
    h5_file: pathlib.Path,
    xdmf_file: pathlib.Path,
    mesh_name: str,
    function_name: str,
    family: str,
    degree: int,
    function_name_vec: str,
) -> dolfin.Function:
    mesh = dolfin.UnitCubeMesh(1, 1, 1)
    V = dolfin.FunctionSpace(mesh, family, degree)
    W = dolfin.VectorFunctionSpace(mesh, family, degree)
    x = dolfin.SpatialCoordinate(mesh)

    f0 = ufl.conditional(ufl.gt(x[0], 0.5), x[1], 2 * x[0])
    v0 = dolfin.project(f0, V)
    w0 = dolfin.interpolate(dolfin.Expression(("x[0]", "3*x[2]", "7*x[1]"), degree=1), W)

    v1 = dolfin.interpolate(dolfin.Expression("x[0]", degree=1), V)
    w1 = dolfin.interpolate(dolfin.Expression(("x[0]", "0", "x[1]"), degree=1), W)

    tt = dolfin.Function(dolfin.FunctionSpace(mesh, "DG", 0))
    tt.interpolate(dolfin.Expression(("x[0]+x[1]"), degree=0))
    ff = dolfin.MeshFunction("size_t", mesh, mesh.topology().dim() - 1, 1)

    class Boundary(dolfin.SubDomain):
        def inside(self, x, on_boundary):
            return on_boundary

    Boundary().mark(ff, 2)

    with dolfin.HDF5File(mesh.mpi_comm(), str(h5_file), "w") as hdf:
        hdf.write(mesh, mesh_name)
        hdf.write(v0, function_name)
        hdf.write(w0, function_name_vec)
        hdf.write(tt, function_name + "DG")
        hdf.write(ff, "facet_tags")
    with dolfin.XDMFFile(mesh.mpi_comm(), str(xdmf_file)) as xdmf:
        xdmf.write(mesh)
        xdmf.write_checkpoint(v0, function_name, 0, dolfin.XDMFFile.Encoding.HDF5, append=True)
        xdmf.write_checkpoint(w0, function_name_vec, 0, dolfin.XDMFFile.Encoding.HDF5, append=True)
        xdmf.write_checkpoint(v1, function_name, 1, dolfin.XDMFFile.Encoding.HDF5, append=True)
        xdmf.write_checkpoint(w1, function_name_vec, 1, dolfin.XDMFFile.Encoding.HDF5, append=True)
        # Legacy DG-0 checkpoint has to be scalar
        xdmf.write_checkpoint(
            tt, function_name + "DG_checkpoint", 0.5, dolfin.XDMFFile.Encoding.HDF5, append=True
        )

    return v0, w0, v1, w1


def verify_hdf5(
    v_ref: dolfin.Function,
    w_ref: dolfin.Function,
    h5_file: pathlib.Path,
    mesh_name: str,
    function_name: str,
    family: str,
    degree: int,
    function_name_vec: str,
):
    mesh = dolfin.Mesh()
    with dolfin.HDF5File(mesh.mpi_comm(), str(h5_file), "r") as hdf:
        hdf.read(mesh, mesh_name, False)
        V = dolfin.FunctionSpace(mesh, family, degree)
        v = dolfin.Function(V)
        hdf.read(v, function_name)

        W = dolfin.VectorFunctionSpace(mesh, family, degree)
        w = dolfin.Function(W)
        hdf.read(w, function_name_vec)

    assert np.allclose(v.vector().get_local(), v_ref.vector().get_local())
    assert np.allclose(w.vector().get_local(), w_ref.vector().get_local())


def verify_xdmf(
    v0_ref: dolfin.Function,
    w0_ref: dolfin.Function,
    v1_ref: dolfin.Function,
    w1_ref: dolfin.Function,
    xdmf_file: pathlib.Path,
    function_name: str,
    family: str,
    degree: int,
    function_name_vec: str,
):
    mesh = dolfin.Mesh()
    with dolfin.XDMFFile(mesh.mpi_comm(), str(xdmf_file)) as xdmf:
        xdmf.read(mesh)
        V = dolfin.FunctionSpace(mesh, family, degree)
        v0 = dolfin.Function(V)
        xdmf.read_checkpoint(v0, function_name, 0)
        v1 = dolfin.Function(V)
        xdmf.read_checkpoint(v1, function_name, 1)

        W = dolfin.VectorFunctionSpace(mesh, family, degree)
        w0 = dolfin.Function(W)
        xdmf.read_checkpoint(w0, function_name_vec, 0)
        w1 = dolfin.Function(W)
        xdmf.read_checkpoint(w1, function_name_vec, 1)

    assert np.allclose(v0.vector().get_local(), v0_ref.vector().get_local())
    assert np.allclose(w0.vector().get_local(), w0_ref.vector().get_local())

    assert np.allclose(v1.vector().get_local(), v1_ref.vector().get_local())
    assert np.allclose(w1.vector().get_local(), w1_ref.vector().get_local())


def create_reference_P1_data(filename: pathlib.Path, mesh_name: str, function_name: str):
    mesh = dolfin.RectangleMesh(dolfin.Point(0.1, 0.2), dolfin.Point(2, 3), 12, 13)
    mesh.rename(mesh_name, mesh_name)
    W = dolfin.VectorFunctionSpace(mesh, "Lagrange", 1)
    w = dolfin.Function(W, name=function_name)
    w.interpolate(dolfin.Expression(("x[0]", "x[1]-x[0]"), degree=1))

    dolfin.VTKFile(str(filename.with_suffix(".vtu")), "ascii").write(mesh)
    dolfin.File(str(filename.with_name(filename.name + "_func").with_suffix(".pvd"))) << w


if __name__ == "__main__":
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--family", type=str, default="DG")
    parser.add_argument("--degree", type=int, default=2)
    parser.add_argument("--output-dir", type=str, default="legacy", dest="dir")
    parser.add_argument("--mesh-name", type=str, default="mesh", dest="name")
    parser.add_argument("--function-name", type=str, default="v", dest="f_name")
    parser.add_argument("--function-name-vec", type=str, default="w", dest="f_name_vec")

    inputs = parser.parse_args()
    path = pathlib.Path(inputs.dir)
    path.mkdir(exist_ok=True, parents=True)
    h5_filename = path / f"{inputs.name}.h5"
    xdmf_filename = path / f"{inputs.name}_checkpoint.xdmf"
    p1_filename = path / f"{inputs.name}_P1"

    v0_ref, w0_ref, v1_ref, w1_ref = create_reference_data(
        h5_filename,
        xdmf_filename,
        inputs.name,
        inputs.f_name,
        inputs.family,
        inputs.degree,
        inputs.f_name_vec,
    )

    verify_hdf5(
        v0_ref,
        w0_ref,
        h5_filename,
        inputs.name,
        inputs.f_name,
        inputs.family,
        inputs.degree,
        inputs.f_name_vec,
    )

    verify_xdmf(
        v0_ref,
        w0_ref,
        v1_ref,
        w1_ref,
        xdmf_filename,
        inputs.f_name,
        inputs.family,
        inputs.degree,
        inputs.f_name_vec,
    )

    P1_ref = create_reference_P1_data(
        filename=p1_filename, mesh_name=inputs.name, function_name=inputs.f_name_vec
    )



# # Time-dependent mesh checkpoints
# As we have seen in the previous examples, we store information about the connectivity,
# the coordinates of the mesh nodes,
# as well as a reference element. Note that the only thing that can change for a mesh
# during a simulation are the coordinate of the mesh nodes.
# In the following example, we will demonstrate how to write a time-dependent mesh
# checkpoint to disk.

# First, we create a simple function to compute the volume of a mesh

# +
import logging
from pathlib import Path

from mpi4py import MPI

import ipyparallel as ipp

import io4dolfinx


def compute_volume(mesh, time_stamp):
    from mpi4py import MPI

    import dolfinx
    import ufl

    # Compute the volume of the mesh
    vol_form = dolfinx.fem.form(1 * ufl.dx(domain=mesh))
    vol_local = dolfinx.fem.assemble_scalar(vol_form)
    vol_glob = mesh.comm.allreduce(vol_local, op=MPI.SUM)
    if mesh.comm.rank == 0:
        print(f"{mesh.comm.rank + 1}/{mesh.comm.size} Time: {time_stamp} Mesh Volume: {vol_glob}")


def write_meshes(filename: Path):
    from mpi4py import MPI

    import dolfinx
    import numpy as np

    import io4dolfinx

    # Create a unit cube
    mesh = dolfinx.mesh.create_unit_cube(
        MPI.COMM_WORLD,
        3,
        6,
        5,
        cell_type=dolfinx.mesh.CellType.hexahedron,
        ghost_mode=dolfinx.mesh.GhostMode.shared_facet,
    )

    # Write mesh to file, associated with time stamp 1.5
    io4dolfinx.write_mesh(filename, mesh, time=1.5)
    compute_volume(mesh, 1.5)
    mesh.geometry.x[:, 0] += 0.1 * mesh.geometry.x[:, 0]
    mesh.geometry.x[:, 1] += 0.3 * mesh.geometry.x[:, 1] * np.sin(mesh.geometry.x[:, 2])
    compute_volume(mesh, 3.3)
    # Write mesh to file, associated with time stamp 3.3
    # Note that we set the mode to append, as we have already created the file
    # and we do not want to overwrite the existing data
    io4dolfinx.write_mesh(filename, mesh, time=3.3, mode=io4dolfinx.FileMode.append)


# -

# We write the sequence of meshes to file

# +
mesh_file = Path("timedep_mesh.bp")
n = 3

with ipp.Cluster(engines="mpi", n=n, log_level=logging.ERROR) as cluster:
    # Write mesh to file
    cluster[:].push({"compute_volume": compute_volume})
    query = cluster[:].apply_async(write_meshes, mesh_file)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))
# -

# # Reading a time dependent mesh
# The only thing we need to do to read the mesh is to send in the associated time stamp,
# which we do by adding `time=time_stamp` when calling {py:func}`io4dolfinx.read_mesh`.

second_mesh = io4dolfinx.read_mesh(
    mesh_file, comm=MPI.COMM_WORLD, backend="adios2", backend_args={"engine": "BP4"}, time=3.3
)
compute_volume(second_mesh, 3.3)

first_mesh = io4dolfinx.read_mesh(
    mesh_file, comm=MPI.COMM_WORLD, backend="adios2", backend_args={"engine": "BP4"}, time=1.5
)
compute_volume(first_mesh, 1.5)

# We observe that the volume of the mesh has changed, as we have perturbed the mesh
# between the two time stamps.
# We also note that we can read the meshes in on a different number of processes than
# we wrote them with and in a different order (as long as the time stamps are correct).



# Copyright (C) 2023 Jørgen Schartum Dokken
#
# This file is part of io4dolfinx
#
# SPDX-License-Identifier:    MIT

"""Top-level package for ADIOS2Wrappers."""

from importlib.metadata import metadata

from .backends import FileMode, get_backend
from .checkpointing import (
    read_attributes,
    read_function,
    read_function_names,
    read_mesh,
    read_meshtags,
    read_timestamps,
    write_attributes,
    write_cell_data,
    write_function,
    write_mesh,
    write_meshtags,
    write_point_data,
)
from .original_checkpoint import write_function_on_input_mesh, write_mesh_input_order
from .readers import (
    read_cell_data,
    read_function_from_legacy_h5,
    read_mesh_from_legacy_h5,
    read_point_data,
)
from .snapshot import snapshot_checkpoint
from .utils import reconstruct_mesh

meta = metadata("io4dolfinx")
__version__ = meta["Version"]
__author__ = meta.get("Author", "")
__license__ = meta["License"]
__email__ = meta["Author-email"]
__program_name__ = meta["Name"]

__all__ = [
    "FileMode",
    "write_meshtags",
    "read_meshtags",
    "read_cell_data",
    "read_mesh",
    "write_mesh",
    "read_function_from_legacy_h5",
    "read_mesh_from_legacy_h5",
    "write_function",
    "read_function",
    "snapshot_checkpoint",
    "write_function_on_input_mesh",
    "write_mesh_input_order",
    "write_attributes",
    "write_data",
    "read_attributes",
    "read_function_names",
    "read_point_data",
    "read_timestamps",
    "get_backend",
    "write_cell_data",
    "write_point_data",
    "reconstruct_mesh",
]



# # Writing a mesh checkpoint
#
# In this example, we will demonstrate how to write a mesh checkpoint to disk.
#
# We start by creating a simple {py:func}`unit-square mesh<dolfinx.mesh.create_unit_square>`.

# +
import logging
from pathlib import Path

from mpi4py import MPI

import dolfinx
import ipyparallel as ipp

mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)
# -

# Note that when a mesh is created in DOLFINx, we send in a
# {py:class}`MPI communicator<mpi4py.MPI.Comm>`.
# The communicator is used to partition (distribute) the mesh across the available processes.
# This means that each process only have access to a sub-set of cells and nodes of the mesh.
# We can inspect these with the following commands:


def print_mesh_info(mesh: dolfinx.mesh.Mesh):
    cell_map = mesh.topology.index_map(mesh.topology.dim)
    node_map = mesh.geometry.index_map()
    print(
        f"Rank {mesh.comm.rank}: number of owned cells {cell_map.size_local}",
        f", number of ghosted cells {cell_map.num_ghosts}\n",
        f"Number of owned nodes {node_map.size_local}",
        f", number of ghosted nodes {node_map.num_ghosts}",
    )


print_mesh_info(mesh)

# ## Create a distributed mesh
# Next, we can use {py:mod}`IPython parallel<ipyparallel>` to inspect a partitioned
# {py:class}`mesh<dolfinx.mesh.Mesh>`.
# We create a convenience function for creating a mesh that
# {py:attr}`shares cells<dolfinx.mesh.GhostMode.shared_facet>` on the boundary
# between two processes if `ghosted=True`.


def create_distributed_mesh(ghosted: bool, N: int = 10):
    """
    Create a distributed mesh with N x N cells. Share cells on process boundaries
    if ghosted is set to True
    """
    from mpi4py import MPI

    import dolfinx

    ghost_mode = dolfinx.mesh.GhostMode.shared_facet if ghosted else dolfinx.mesh.GhostMode.none
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, N, N, ghost_mode=ghost_mode)
    print(f"{ghost_mode=}")
    print_mesh_info(mesh)


# Next we start up a new cluster with three engines.
# As we defined `print_mesh_info` locally on this process, we need to push it to all engines.

with ipp.Cluster(engines="mpi", n=3, log_level=logging.ERROR) as cluster:
    # Push print_mesh_info to all engines
    cluster[:].push({"print_mesh_info": print_mesh_info})

    # Create mesh with ghosted cells
    query_true = cluster[:].apply_async(create_distributed_mesh, True)
    query_true.wait()
    assert query_true.successful(), query_true.error
    print("".join(query_true.stdout))
    # Create mesh without ghosted cells
    query_false = cluster[:].apply_async(create_distributed_mesh, False)
    query_false.wait()
    assert query_false.successful(), query_false.error
    print("".join(query_false.stdout))

# ## Writing a mesh checkpoint
# The input data to a mesh is:
# - A geometry: the set of points in R^D that are part of each cell
# - A two-dimensional connectivity array: A list that indicates which nodes of the geometry
#    is part of each cell
# - A {py:func}`reference element<basix.ufl.element>`: Used for push data back and
#   forth from the reference element and computing Jacobians
# We now use {py:mod}`io4dolfinx` to write a mesh to file.


def write_mesh(filename: Path):
    import subprocess

    from mpi4py import MPI

    import dolfinx

    import io4dolfinx

    # Create a simple unit square mesh
    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD, 10, 10, cell_type=dolfinx.mesh.CellType.quadrilateral
    )

    # Write mesh checkpoint
    io4dolfinx.write_mesh(filename, mesh, backend="adios2", backend_args={"engine": "BP4"})

    # Inspect checkpoint on rank 0 with `bpls`
    if mesh.comm.rank == 0:
        output = subprocess.run(["bpls", "-a", "-l", str(filename.absolute())], capture_output=True)
        print(output.stdout.decode("utf-8"))


# +
mesh_file = Path("mesh.bp")

with ipp.Cluster(engines="mpi", n=2, log_level=logging.ERROR) as cluster:
    # Write mesh to file
    query = cluster[:].apply_async(write_mesh, mesh_file)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))
# -

# We observe that we have stored all the data needed to re-create the mesh in the file `mesh.bp`.
# We can therefore read it (to any number of processes) with {py:func}`io4dolfinx.read_mesh`


def read_mesh(filename: Path):
    from mpi4py import MPI

    import dolfinx

    import io4dolfinx

    mesh = io4dolfinx.read_mesh(
        filename,
        comm=MPI.COMM_WORLD,
        backend="adios2",
        backend_args={"engine": "BP4"},
        ghost_mode=dolfinx.mesh.GhostMode.none,
    )
    print_mesh_info(mesh)


# ## Reading mesh checkpoints (N-to-M)
# We can now read the checkpoint on a different number of processes than we wrote it on.

with ipp.Cluster(engines="mpi", n=4, log_level=logging.ERROR) as cluster:
    # Write mesh to file
    cluster[:].push({"print_mesh_info": print_mesh_info})
    query = cluster[:].apply_async(read_mesh, mesh_file)
    query.wait()
    assert query.successful(), query.error
    print("".join(query.stdout))
